#include "stdafx.h"

//#include "Network.h"
#include "FixedPackets.h"

cACServer::cACServer()
{
	//Constructor

	Listening = false;

	WSADATA wsaData;
	WORD wVersionRequested = 0x0202;
	int err;

	err = WSAStartup(wVersionRequested, &wsaData); //for winsock startup

	for (int i=0;i<128;i++)
	{
		ConnUser[i].Connected = false;
		ConnUser[i].State = 0;
		ConnUser[i].Char = 0;
	}

	MaxUsers = 0;
	NumConnected = 0;

	NumStored = 0;
	for (i=0;i<16384;i++)
		IPStore[i].Enabled = false;

	NeedToDelete = true;

	LoadOptions();

//	CalcNextFreeGUID();
}

cACServer::~cACServer()
{
	//Destructor
	SaveObjects();

	for (int i=0;i<128;i++)
		if (ConnUser[i].Connected)
			DisconnectUser(i);

	WSACleanup(); // Always do this at the end of a winsock program
}

void cACServer::LoadOptions()
{
	HKEY keyout;
	RegCreateKey(HKEY_CURRENT_USER, "Software\\ACServer\\Server", &keyout);

	long outvallen;
	outvallen = 2048;
	if (RegQueryValue(keyout, "MOTD", MOTDb, &outvallen) != ERROR_SUCCESS)
	{
		sprintf(MOTDb, "This server's operator needs to set its MOTD in the options...\n");
	}
	outvallen = 64;
	if (RegQueryValue(keyout, "ServerName", ServerName, &outvallen) != ERROR_SUCCESS)
	{
		sprintf(ServerName, "Unnamed Server");
	}
	outvallen = 4;
	if (RegQueryValueEx(keyout, "ListenPort", 0, NULL, (BYTE *) &ListenPort, (DWORD *) &outvallen) != ERROR_SUCCESS)
	{
		ListenPort = 9002;
	}
	
	RegCloseKey(keyout);
}

void cACServer::SaveOptions()
{
	HKEY keyout;
	RegCreateKey(HKEY_CURRENT_USER, "Software\\ACServer\\Server", &keyout);

	RegSetValue(keyout, "MOTD", REG_SZ, MOTDb, strlen(MOTDb));
	RegSetValue(keyout, "ServerName", REG_SZ, ServerName, strlen(ServerName));
	RegSetValueEx(keyout, "ListenPort", 0, REG_DWORD, (const BYTE *) &ListenPort, 4);
	
	RegCloseKey(keyout);
}

void cACServer::StartListen()
{
	LoadObjects();

	Listening = true;

	srand(time(NULL));

	struct sockaddr_in localAddress;	//IP and port of the local machine
	localAddress.sin_family = AF_INET;
	localAddress.sin_addr.s_addr = inet_addr(LocalIP);
	localAddress.sin_port = htons(ListenPort);

	charSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	bind(charSocket, (struct sockaddr *)&localAddress, sizeof(sockaddr_in) );

	listen(charSocket, SOMAXCONN);

	sprintf(sayit, "ACServer Server Up!  Listening on port %i...", ListenPort); LogDisp(sayit);

	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_RESETCONTENT, 0, 0);
	char tosay[100];
	sprintf(tosay, "User\tIP:Port\tStatus"); SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_ADDSTRING, 0, (WPARAM) &tosay[0]);
	for (int i=0;i<128;i++)
	{
		sprintf(tosay, "%i\tN/A\tNot Connected", i);
		SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_ADDSTRING, 0, (WPARAM) &tosay[0]);
	}

	//Set up event notification on main listen socket
	int err = WSAAsyncSelect(charSocket, MainWnd, WM_USER+1, FD_ACCEPT/* | FD_READ*/);
}

void cACServer::LoadAccounts()
{	
	FILE *in = fopen("Password.dat","rb");
	if (!in)
	{
		NumPasses = 0;
		return;
	}
	
	fseek(in, 0, SEEK_END);
	int filelen = ftell(in);
	fseek(in, 0, SEEK_SET);

	NumPasses = (int) (filelen/sizeof(stAccounts));

	fread(ZonePasses, filelen, 1, in);

	for (int i=0;i<NumPasses;i++)
	{
		ZonePasses[i].LoggedIn = (WORD) -1;
	}

	fclose(in);
}

void cACServer::CalcNextFreeGUID()
{
	NextFreeCharGUID = 0x50000000;
	for (int i=0;i<NumPasses;i++)
	{
		for (int j=0;j<5;j++)
		{
			if (ZonePasses[i].Chars[j] >= NextFreeCharGUID)
				NextFreeCharGUID = ZonePasses[i].Chars[j] + 1;
		}
	}
}

void cACServer::SaveAccounts()
{
	FILE *out = fopen("Password.dat","wb");
	if (!out)
	{
		MessageBox(NULL, "Error on Password File Save!  Very bad!!", "Critical Error!", MB_OK);
		return;
	}

	fwrite(ZonePasses, sizeof(stAccounts)*NumPasses, 1, out);

	fclose(out);
}


void cACServer::CheckMessages(int UserNum, short EventType)
{
	unsigned int amountRead=0;
	int size=sizeof(sockaddr_in);
	cClientPacketHeader tphdr;

	if (UserNum == 0)
	{
		sockaddr_in tpsa; int tpsal = sizeof(tpsa);
		SOCKET tpsck = accept(charSocket, (sockaddr *) &tpsa, &tpsal);
		int newbuf = 10000;
		setsockopt(tpsck, SOL_SOCKET, SO_SNDBUF, (const char *) &newbuf, sizeof(newbuf));

		if (tpsck != SOCKET_ERROR)
		{
			for (int i=0;i<128;i++)
			{
				if (!ConnUser[i].Connected)
				{
					memcpy(&ConnUser[i].sockaddy,&tpsa,sizeof(sockaddr));
					ConnUser[i].Socket = tpsck;

					ConnUser[i].Connected = true;

					ConnUser[i].State = -1;

					int err = WSAAsyncSelect(tpsck, MainWnd, WM_USER+2+i, FD_READ);

					break;
				}
			}
		}

		return;
	}

	if (EventType != FD_READ)
	{
		return;
	}
		
	int UserParsing = UserNum - 1;

	amountRead=::recv(ConnUser[UserParsing].Socket,(char *) TestBuf,4096,NULL);

	if (ConnUser[UserParsing].State == -1)
	{
		//Get l/p here...
		char tpLogin[20], tpPass[20];
		memcpy(tpLogin, TestBuf, 20);
		memcpy(tpPass, TestBuf+20, 20);
		
		bool LoginFound = false;

		int LoginRet = 2;
		//0 = Valid Existing Login
		//1 = Valid New Login
		//2 = Incorrect Password

		//Check for existing Login by that name, or else create new one...
		for (int h=0;h<NumPasses;h++)
		{
			if (strcmp(ZonePasses[h].Login,tpLogin) == 0)
			{
				//ID Found
				LoginFound = true;
				if (strcmp(ZonePasses[h].Pass,tpPass) == 0)
				{
					//Valid Password

					sprintf(sayit, "User %s logged in.", ZonePasses[h].Login); LogDisp(sayit);
					LoginRet = 0;

					ConnUser[UserParsing].ZoneNum = h;
				}
				h = NumPasses;
			}
		}
		
		if (!LoginFound)
		{
			memcpy(ZonePasses[NumPasses].Login, tpLogin, 20);
			memcpy(ZonePasses[NumPasses].Pass, tpPass, 20);
			ZeroMemory(ZonePasses[NumPasses].Chars, sizeof(ZonePasses[NumPasses].Chars));
			ZeroMemory(ZonePasses[NumPasses].Secs2Del, sizeof(ZonePasses[NumPasses].Secs2Del));
			ZeroMemory(ZonePasses[NumPasses].Names, sizeof(ZonePasses[NumPasses].Names));
			sprintf(sayit, "New User %s logged in.", ZonePasses[NumPasses].Login); LogDisp(sayit);
			LoginRet = 1;
			ZonePasses[NumPasses].LoggedIn = -1;

			ConnUser[UserParsing].ZoneNum = NumPasses;

			NumPasses++;
			SaveAccounts();
		}

		memcpy(ConnUser[UserParsing].ZoneName, ZonePasses[ConnUser[UserParsing].ZoneNum].Login, 20);
		
		//Return LoginRet here...
		//
		//!!!
		send(ConnUser[UserParsing].Socket, (char *) &LoginRet, 4, NULL);

		if (LoginRet == 2)
		{
			//Invalid Login...

			ConnUser[UserParsing].Connected = false;

			closesocket(ConnUser[UserParsing].Socket);

			return;
		}					

		ConnUser[UserParsing].State = 1;
		
		ConnUser[UserParsing].curSeq = 1;
		ConnUser[UserParsing].EventCount = 0;

		ConnUser[UserParsing].msgID = 1;

		ConnUser[UserParsing].bTime = 0;
		ConnUser[UserParsing].RecTime = 0;
		ConnUser[UserParsing].LastTime = 0;

		ConnUser[UserParsing].move_count = 1;

		ConnUser[UserParsing].SelectedItem = 0;

		ConnUser[UserParsing].PortalCount = 0;
		ConnUser[UserParsing].OverrideCount = 0;
		ConnUser[UserParsing].AnimCount = 0;
		ConnUser[UserParsing].MoveItemCount = 1;
		ConnUser[UserParsing].EquipCount = 0;
		ConnUser[UserParsing].AttackCount = 0;
		ZeroMemory(ConnUser[UserParsing].Count237, sizeof(ConnUser[UserParsing].Count237));
		ZeroMemory(ConnUser[UserParsing].Count244,3);
		
		ConnUser[UserParsing].Connable = true;
		ConnUser[UserParsing].ConnTimer = 0;

		UpdateCharInfo(UserParsing, "Connected...");

		if (UserParsing >= MaxUsers)
			MaxUsers = UserParsing + 1;

		return;
	}

	memcpy(&tphdr, &TestBuf[0], sizeof(tphdr));

	ConnUser[UserParsing].LastTime = ConnUser[UserParsing].RecTime;
	ConnUser[UserParsing].lastInSeq = tphdr.m_dwSequence;

//	if (tphdr.m_dwFlags == 0x0000000A)
//	{	sprintf(sayit, "* Error Packet - A..."); LogDisp(sayit); return; }
	if (tphdr.m_dwFlags == 0x00000002)
	{	sprintf(sayit, "* Error Packet - 2..."); LogDisp(sayit); return; }

	if (tphdr.m_dwFlags == 0x00000004)
	{
		if (ConnUser[UserParsing].bTime < tphdr.m_bTime)
			ConnUser[UserParsing].bTime = tphdr.m_bTime;

		//reply in some way to this at some point...
		
		return;
	}

	if (tphdr.m_dwFlags == 0x00100000)
		return;

	if (tphdr.m_dwFlags == 0x00000020)
	{
		DisconnectUser(UserParsing);
		return;
	}

	if (ConnUser[UserParsing].State == 1)
	{
		//Character Selection Screen, Just Pings...  Then Entering Game.
		if (tphdr.m_dwFlags == 0x00000200)
		{
			DWORD MessageType;	memcpy(&MessageType, &TestBuf[sizeof(cClientPacketHeader)], 4);
			
			switch (MessageType)
			{
			case 0xF656:
				{
					//Character Creation!!
					DWORD beefbeef, tpdword;
					BYTE shit[0x170], *PackPointer;
					PackPointer = &TestBuf[sizeof(cClientPacketHeader) + 4];
					memcpy(&beefbeef, PackPointer, 4);	PackPointer += 4;
					memcpy(shit, PackPointer, 0x168);	PackPointer += 0x168;	//starts at $20
					
					WORD CharNameLen;
					BYTE CharName[32];
					memcpy(&CharNameLen, PackPointer, 2);			PackPointer += 2;
					memcpy(CharName, PackPointer, CharNameLen);		PackPointer += CharNameLen;

					DWORD CharNum;
					for (int i=0;i<5;i++)
					{
						if (!ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i])
						{
							memcpy(ZonePasses[ConnUser[UserParsing].ZoneNum].Names[i], CharName,32);
							CalcNextFreeGUID();
							ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] = NextFreeCharGUID;
							CharNum = i;
							i = 5;
	
							SaveAccounts();
						}
					}

					//Fill out stCharType and send to WorldServer
					cWorldObject_Char * NewChar = new cWorldObject_Char();
					NewChar->GUID = ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[CharNum];
					memcpy(NewChar->Name, CharName, 32);
					
					//Read in shit
					PackPointer = &shit[0];
//					char ZN[20];		memcpy(ZN, PackPointer, 20);	PackPointer += 20;				//Unknown1
//					DWORD unk1;			memcpy(&unk1, PackPointer, 4);	PackPointer += 4;				//Unknown1
//										memcpy(&NewChar->Gender , PackPointer, 4);	PackPointer += 4;	//Guessing its Gender
//										memcpy(&NewChar->Race , PackPointer, 4);	PackPointer += 4;	//Race, Aluv/Gharu/Sho 0/1/2
					/*$70 unknown right now*/									PackPointer += 0xA4;
										memcpy(&NewChar->InitialStat[0], PackPointer, 6*4);	PackPointer += 6*4;
					/*$10 unknown right now*/									PackPointer += 0x10;
					

					sprintf(sayit, "* %s created a new char, %s with GUID: %08X", ConnUser[UserParsing].ZoneName, NewChar->Name , NewChar->GUID); LogDisp(sayit);

					ZeroMemory(NewChar->SkillInc, sizeof(NewChar->SkillInc));
					ZeroMemory(NewChar->SkillFreeXP, sizeof(NewChar->SkillFreeXP));
					ZeroMemory(NewChar->SkillXP, sizeof(NewChar->SkillXP));
					DWORD tpskill;
					for (i=0; i < 39; i++)
					{
						memcpy(&tpskill, PackPointer, 4);
						PackPointer += 4;
						NewChar->SkillTrain[i+1] = 0;
						if (tpskill)
						{
							NewChar->SkillTrain[i+1] = tpskill;
							if (tpskill == 1)
							{
								NewChar->SkillInc[i+1] = 0;
								NewChar->SkillFreeXP[i+1] = 0;
							}
							if (tpskill == 2)
							{
								NewChar->SkillInc[i+1] = 5;
								NewChar->SkillFreeXP[i+1] = 526;
							}
							if (tpskill == 3)
							{
								NewChar->SkillInc[i+1] = 10;
								NewChar->SkillFreeXP[i+1] = 671;
							}

							NewChar->SkillXP[i+1] = 0;
						}
					}
					
					NewChar->Vitae = 0;

					NewChar->Level = 1;

					//Eventually decode starting place... For now teth works. :)
					NewChar->Loc.landblock = 0x25810034;
					NewChar->Loc.x = 156.1f;
					NewChar->Loc.y = 81.9f;
					NewChar->Loc.z = 220.0f;

					tpdword = 0x3F579630; memcpy(&NewChar->Loc.a, &tpdword, 4);
					NewChar->Loc.b = 0x00000000;
					NewChar->Loc.c = 0x00000000;
					NewChar->Loc.w = 90.1f;	//Heading

					memcpy(&NewChar->LSTie, &NewChar->Loc, sizeof(NewChar->Loc));

					NewChar->CurrentSecondaryStat[0] = NewChar->InitialStat[1]/2;
					NewChar->CurrentSecondaryStat[1] = NewChar->InitialStat[1];
					NewChar->CurrentSecondaryStat[2] = NewChar->InitialStat[5];

					ZeroMemory(NewChar->XPIntoSecondaryStat, sizeof(NewChar->XPIntoSecondaryStat));
					ZeroMemory(NewChar->XPIntoStat, sizeof(NewChar->XPIntoStat));
					ZeroMemory(NewChar->SecondaryStatInc, sizeof(NewChar->SecondaryStatInc));

					memcpy(NewChar->CurrentStat, NewChar->InitialStat, 6*4);

					ZeroMemory(NewChar->MainPack, sizeof(NewChar->MainPack));
					ZeroMemory(NewChar->SidePacks, sizeof(NewChar->SidePacks));
					ZeroMemory(NewChar->Equipped, sizeof(NewChar->Equipped));
					NewChar->Model = 1;
					NewChar->Monarch = 0;
					NewChar->NumLogins = 1;
					NewChar->Options = 0x01E8E543;
					NewChar->Patron = 0;
					NewChar->SkillCredits = 0;
					ZeroMemory(NewChar->SpellsLearned, sizeof(NewChar->SpellsLearned));
					NewChar->TotalXP = 0;
					NewChar->UnassignedXP = 0;
					NewChar->Vitae = 0;
					NewChar->Class = 0;
					ZeroMemory(NewChar->Vassal, sizeof(NewChar->Vassal));
					ZeroMemory(NewChar->ShortCut, sizeof(NewChar->ShortCut));
					ZeroMemory(NewChar->SpellsLearned, sizeof(NewChar->SpellsLearned));
					ZeroMemory(NewChar->SpellBar, sizeof(NewChar->SpellBar));

					FILE *out = fopen("CharList.dat","a+b");

					fseek(out, 0, SEEK_END);
					
					BYTE TPlog[0x2000]; int charsize = NewChar->Serialize(TPlog);
					fwrite(&NewChar->GUID, 4, 1, out);
					fwrite(NewChar->Name, 32, 1, out);
					fwrite(&NewChar->Loc, sizeof(Location_t), 1, out);
					fwrite(TPlog, charsize, 1, out);

					fclose(out);

					sprintf(sayit, "New char created: %s with GUID %08X", NewChar->Name, NewChar->GUID); LogDisp(sayit);

					//Create Char Respond Packet...
					BYTE CreateChar[0x100];
					
					
					PackPointer = &CreateChar[0];

					tpdword = 0x0000F643;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
					tpdword = 0x00000001;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;	//Success
					tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[CharNum];	memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
					int outlen; GenString(ZonePasses[ConnUser[UserParsing].ZoneNum].Names[CharNum], &outlen, PackPointer); PackPointer += outlen;
					tpdword = 0x00000000;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;

					DWORD PackLen = (DWORD) ((DWORD) PackPointer - (DWORD) CreateChar);

					Send200(CreateChar, PackLen, 4, UserParsing);

					//Resend Char List!
					SendCharList(UserParsing);
				}
				break;
			case 0xF7C8:
				{
					//Entering Game
					UpdateCharInfo(UserParsing, "Got Enter Game Packet...");

					//Now to return a fun 200 packet... whee...
					DWORD tph3 = 0x0000F7C7;

					Send200((BYTE *) &tph3, sizeof(tph3), 4, UserParsing);

					ConnUser[UserParsing].State = 2;
				}
				break;
			case 0xF7D9:
				//Restoring Character
				{
					DWORD GUIDtoRestore; DWORD Ones[2];
					memcpy(&GUIDtoRestore, &TestBuf[sizeof(tphdr)+4],4);
					memcpy(&Ones[0], &TestBuf[sizeof(tphdr)+8],4);
					memcpy(&Ones[1], &TestBuf[sizeof(tphdr)+12],4);

					int CharRestoring = 0;
					for (int i=0;i<5;i++)
					{
						if ((ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x7FFFFFFF) == GUIDtoRestore)
						{
							ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] &= 0x7FFFFFFF;
							
							SaveAccounts();

							CharRestoring = i;
							i = 5;
						}
					}

					BYTE CreateChar[0x100], *PackPointer = &CreateChar[0];
					DWORD tpdword;

					tpdword = 0x0000F643;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
					tpdword = 0x00000001;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
					tpdword = GUIDtoRestore;	memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;

					tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[CharRestoring] & 0x7FFFFFFF;
						memcpy(PackPointer,&tpdword,4); PackPointer += 4;	//GUID

					BYTE NameLen;
					NameLen = strlen(ZonePasses[ConnUser[UserParsing].ZoneNum].Names[CharRestoring]) + 1;
						memcpy(PackPointer,&NameLen,sizeof(NameLen)); PackPointer += sizeof(NameLen);	//NameLen

					*PackPointer = 0; PackPointer++;												//0
					NameLen--; memcpy(PackPointer,&ZonePasses[ConnUser[UserParsing].ZoneNum].Names[CharRestoring],NameLen); PackPointer += NameLen;	//Name
					*PackPointer = 0; PackPointer++;												//0
					
					int Tpl = ((PackPointer-&CreateChar[0])%4); if (Tpl) Tpl=4-Tpl;			//Padding
					for (int ii=0;ii<Tpl;ii++) {	*PackPointer = 0; PackPointer++; }		//"
					
					tpdword = 0x00000000;		memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
					Send200(&CreateChar[0], (int) (PackPointer - &CreateChar[0]), 4, UserParsing);

					//Resend Char List!
					SendCharList(UserParsing);
				}
				break;
			case 0xF655:
				//Deleting Character
				{
					DWORD beefbeef; char ZoneName[20]; DWORD LastWord[2];
					memcpy(&beefbeef, &TestBuf[sizeof(tphdr)+4],4);
					memcpy(&ZoneName, &TestBuf[sizeof(tphdr)+8],20);
					memcpy(&LastWord[0], &TestBuf[sizeof(tphdr)+28],4);
					memcpy(&LastWord[1], &TestBuf[sizeof(tphdr)+32],4);

					//LastWord[0] seems to be 01f7 all the time
					//LastWord[1] is the index of the char deleted, in order of chars sent, i think...
					
					ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[LastWord[1]] |= 0x80000000;	//To be deleted
					ZonePasses[ConnUser[UserParsing].ZoneNum].Secs2Del[LastWord[1]] = 300;	//5 Minutes
					SaveAccounts();
					
					NeedToDelete = true;

					//Resend Char List!
					SendCharList(UserParsing);
				}
				break;
			default:
				break;
			}
		} else if ((tphdr.m_dwFlags == 0x000080) && (ConnUser[UserParsing].Connable)) {
//				sprintf(sayit, "-=*=- Connection Packet Detected... Logging In... -=*=-"); LogDisp(sayit);

			UpdateCharInfo(UserParsing, "Sending Initial Info...");
			
			ConnUser[UserParsing].Connable = false;
			ConnUser[UserParsing].ConnTimer = 5;

			SendCharList(UserParsing);

			UpdateCharInfo(UserParsing, "Selecting Character");

			return;
		} else if (tphdr.m_dwFlags == 0x00000022) {
			//Ping...
		} else if (tphdr.m_dwFlags == 0x00008002) {
			//Proxy Wakeup
		} else if (tphdr.m_dwFlags == 0x00000002) {
			//Meh
		} else if (tphdr.m_dwFlags == 0x00000202) {
			//CD login-packet
		} else if (tphdr.m_dwFlags == 0x00000102) {
			//Dunno
		} else if (tphdr.m_dwFlags == 0x00001002) {
			ConnUser[UserParsing].bTime++;
			//Btime raised...
		} else if (tphdr.m_dwFlags == 0x00000042) {
			//Dunno...
		} else if (tphdr.m_dwFlags == 0x0000000a) {
			//Error packet...
		} else {
			int aa=4;
			//Dunno...
		}
	} else if (ConnUser[UserParsing].State == 2) {
		//Waiting for Login 200-Ack...
		if (tphdr.m_dwFlags == 0x00000200) {	//f657
			UpdateCharInfo(UserParsing, "Got Enter game request...");

			memcpy(&ConnUser[UserParsing].charGUID, &TestBuf[4+sizeof(cClientPacketHeader)], 4);
			
			ConnUser[UserParsing].State = 3;
		} else if (tphdr.m_dwFlags == 0x00001002) {
			ConnUser[UserParsing].bTime++;
			//Btime raised...
		} else {
			//Dunno...
		}
	} else if (ConnUser[UserParsing].State == 3) {
		//Referring to WorldServer
		if (tphdr.m_dwFlags == 0x00000200) {	//f7b1/001f
			ConnUser[UserParsing].State = 4;

			DWORD GUIDtoLoad = ConnUser[UserParsing].charGUID;

			ServerMessage(UserParsing, "Welcome to KillaServ v-9.432", COLOR_BLUE);
			ServerMessage(UserParsing, "* Type !help for a Features List!", COLOR_BLUE);

			//Load character...
			FILE *charread = fopen("CharList.dat","rb");

			if (!charread)
			{
				//doh
				sprintf(sayit, "Couldn't find Character File... Bad... Disconnecting User..."); LogDisp(sayit);
				
				return;			//Bad...
			}

			fseek(charread, 0, SEEK_END);
			int numchars = ftell(charread)/(SIZEOFCHAR+68);
			fseek(charread, 0, SEEK_SET);

			DWORD GUIDtemp; bool CharFound = false;
			for (int j=0;j<numchars;j++)
			{
				fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);
				fread(&GUIDtemp, 4, 1, charread);
				if (GUIDtemp == GUIDtoLoad)
				{
					fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);
					cWorldObject_Char *tpchar = new cWorldObject_Char();
					ObjectList[GUIDtemp] = tpchar;
					ConnUser[UserParsing].Char = tpchar;

					fread(&tpchar->GUID, 4, 1, charread);
					fread(tpchar->Name, 32, 1, charread);
					fread(&tpchar->Loc, sizeof(Location_t), 1, charread);
					BYTE tpl[5000];
					fread(tpl, SIZEOFCHAR, 1, charread);
					tpchar->LoadFrom(tpl);
					
					j = numchars;
					CharFound = true;
				}
			}

			fclose(charread);

			if (!CharFound)
			{
				sprintf(sayit, "Couldn't find Character's Entry... Bad... Disconnecting User..."); LogDisp(sayit);
				
				return;			//Bad...
			}

			LBObjects[ConnUser[UserParsing].Char->Loc.landblock >> 16][ConnUser[UserParsing].Char->GUID] = (cWorldObject *) ConnUser[UserParsing].Char;

			//Load Objects out of store...
			for (j=0;j<25;j++) {
				if (ConnUser[UserParsing].Char->Equipped[j]) {
					ConnUser[UserParsing].EquippedPtr[j] = ObjectList[ConnUser[UserParsing].Char->Equipped[j]];
					ConnUser[UserParsing].ObjectCache[ConnUser[UserParsing].Char->Equipped[j]] = ConnUser[UserParsing].EquippedPtr[j];
				} else j = 25;
			}
			for (j=0;j<7;j++) {
				if (ConnUser[UserParsing].Char->SidePacks[j]) {
					ConnUser[UserParsing].SidePacksPtr[j] = ObjectList[ConnUser[UserParsing].Char->SidePacks[j]];
					ConnUser[UserParsing].ObjectCache[ConnUser[UserParsing].Char->SidePacks[j]] = ConnUser[UserParsing].SidePacksPtr[j];
				} else j = 7;
			}
			for (j=0;j<102;j++) {
				if (ConnUser[UserParsing].Char->MainPack[j]) {
					ConnUser[UserParsing].MainPackPtr[j] = ObjectList[ConnUser[UserParsing].Char->MainPack[j]];
					ConnUser[UserParsing].ObjectCache[ConnUser[UserParsing].Char->MainPack[j]] = ConnUser[UserParsing].MainPackPtr[j];
				} else j = 102;
			}

			ConnUser[UserParsing].Char->Owner = UserParsing;

			//Hax0r...?
			ConnUser[UserParsing].Char->NumLogins = 3; //++;

			sprintf(sayit, "* Requested Login, Loaded char %08X...", ConnUser[UserParsing].Char->GUID);	LogDisp(sayit);
			
			if (UserParsing >= MaxUsers)
				MaxUsers = UserParsing+1;

			ConnUser[UserParsing].State = 5;

			UpdateCharInfo(UserParsing, "Sending Main Login Info...");

			//Try to login...
			int packlen;
			BYTE CharData[0x2000];

			//SetPackContents eventually here for any of the 7 packs..

			Sleep(1000);

			//Send giant login packet
			packlen = Generate_F7B0_13(CharData, UserParsing);
			Send200(CharData, packlen, 4, UserParsing);

			Sleep(1000);

			//Create him on other ppl..
			BYTE CreatePlayer[0x2000];
			packlen = ConnUser[UserParsing].Char->GenerateCreatePacket(CreatePlayer);

			for (int iii=0;iii<MaxUsers;iii++)
			{
				if ((ConnUser[iii].Connected) && ((ConnUser[iii].State == 6) || (iii == UserParsing)))
					Send200(&CreatePlayer[0], packlen, 3, iii);
			}

			unsigned char LOGIN_CHAR[8] = {
				0x46, 0xF7, 0x00, 0x00, 0x30, 0x9F, 0x03, 0x50,                                                   //(F__0__P)         - 0000
			};

			memcpy(&LOGIN_CHAR[4],&ConnUser[UserParsing].Char->GUID,4);
			Send200(&LOGIN_CHAR[0], sizeof(LOGIN_CHAR), 3, UserParsing);

			sprintf(sayit, "Player %s Joined!", ConnUser[UserParsing].Char->Name);
			ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);
			
			WORD BaseBlock = ConnUser[UserParsing].Char->Loc.landblock >> 16;
			SendLandblock(UserParsing, BaseBlock);
			SendLandblock(UserParsing, BaseBlock-1);
			SendLandblock(UserParsing, BaseBlock+1);
			SendLandblock(UserParsing, BaseBlock-256);
			SendLandblock(UserParsing, BaseBlock+256);
			SendLandblock(UserParsing, BaseBlock-1-256);
			SendLandblock(UserParsing, BaseBlock+1-256256);
			SendLandblock(UserParsing, BaseBlock-1+256);
			SendLandblock(UserParsing, BaseBlock+1+256);

			unsigned char MSG_00000002[16] = {
				0x4B, 0xF7, 0x00, 0x00, 0x30, 0x9F, 0x03, 0x50, 0x10, 0x04, 0x40, 0x00, 0x03, 0x00, 0x02, 0x00,   //(K__0__P__@_____) - 0000
			};
			
			//Exit Portal Mode
			memcpy(&MSG_00000002[4],&ConnUser[UserParsing].Char->GUID,4);
			Send200(&MSG_00000002[0], sizeof(MSG_00000002), 3, UserParsing);

			ConnUser[UserParsing].State = 6;
		}			
	} else if (ConnUser[UserParsing].State == 6) {
		WorldServerParse(UserParsing, TestBuf, amountRead, &tphdr);
	} else {
		char * tps = new char[1 + (amountRead-20)*5], *tps2 = tps;
		for (unsigned int tpl=0; tpl < (amountRead-20) ;tpl++)
		{
			sprintf(tps2, "0x%02X ", TestBuf[tpl+20]);
			tps2+=5;
		}
		sprintf(sayit, "In:Seq:%08X Flg:%08X Srv:%08X CRC:%08X Size:%04X Tm:%02X Table:%02X* Data: %s", tphdr.m_dwSequence, tphdr.m_dwFlags, tphdr.m_dwServerID, tphdr.m_dwCRC, tphdr.m_wTotalSize, tphdr.m_bTime, tphdr.m_bTable, tps); LogDisp(sayit);
		delete [] tps;
	}
}

void cACServer::LogDisp(char *toDisp)
{
	int numtodisp = 1 + (int) (strlen(toDisp)/90);
	char tosay[91];
	tosay[90] = 0;
	int lognum;
	for (int i=0;i<numtodisp;i++)
	{
		memcpy(&tosay[0],toDisp+(i*90),90);
		lognum = SendDlgItemMessage(MainWnd, IDC_LIST1, LB_ADDSTRING/*LB_INSERTSTRING*/, 0, (LPARAM) tosay);
	}

	SendDlgItemMessage(MainWnd, IDC_LIST1, LB_SETTOPINDEX, lognum-16, 0);
}

void cACServer::IncTimes()
{
	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
		{
			if (!ConnUser[i].Connable)
			{
				ConnUser[i].ConnTimer--;
				if (!ConnUser[i].ConnTimer)
					ConnUser[i].Connable = true;
			}

//			ConnUser[i].bTime++;
//			if (ConnUser[i].bTime > 80)
//				ConnUser[i].bTime = 0;

			ConnUser[i].RecTime++;
			if ((ConnUser[i].RecTime - ConnUser[i].LastTime) > 30)		//If 30 seconds...
			{
				DisconnectUser(i);
//				sprintf(sayit, "* User %i timed out, disconnecting...", i);  LogDisp(sayit);
			}
		}
	}

	if (NeedToDelete)
	{
		NeedToDelete = false;
		for (i=0;i<NumPasses;i++)
		{
			for (int j=0;j<5;j++)
			{
				if (ZonePasses[i].Secs2Del[j])
				{
					ZonePasses[i].Secs2Del[j]--;
					if (!ZonePasses[i].Secs2Del[j])
					{
						ZonePasses[i].Chars[j] = 0;
						ZeroMemory(ZonePasses[i].Names[j], 32);
						if (ZonePasses[i].LoggedIn != -1)
							SendCharList(ZonePasses[i].LoggedIn);
						SaveAccounts();
					} else {
						NeedToDelete = true;
					}
				}

			}
		}
	}
}

void cACServer::SendCharList(int UserParsing)
{
	BYTE PlayerList[0x2000];
	BYTE *PackPointer = PlayerList;

	DWORD tpdword;

	tpdword = 0x0000F658;			memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
	tpdword = 0x00000000;			memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
	
	int NumChars = 0;
	for (int i=0;i<5;i++)
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i])
			NumChars++;
									memcpy(PackPointer, &NumChars, 4);	PackPointer += 4;

	int outlen;
	for (i=0;i<5;i++)
	{
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i])
		{
			tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x7FFFFFFF;
				memcpy(PackPointer,&tpdword,4); PackPointer += 4;	//GUID

			GenString(ZonePasses[ConnUser[UserParsing].ZoneNum].Names[i], &outlen, PackPointer); PackPointer += outlen;

			//Time remaining till perma-delete
			if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x80000000)
				tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Secs2Del[i];
			else
				tpdword = 0;

					memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
		}
	}

	tpdword = 0;
	for (i=0;i<5;i++)
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x80000000)
			tpdword++;
									memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//0 - DWORD
	tpdword = 5;					memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//5 - DWORD
	tpdword = 0xBEEFBEEF;			memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//BEEFBEEF - DWORD
	memcpy(PackPointer,&ConnUser[UserParsing].ZoneName[0],0x14);	PackPointer += 20;		//Zone Name - 20 Chars
	tpdword = 0;					memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//0 - DWORD
	Send200(PlayerList, (int) (PackPointer - PlayerList), 4, UserParsing);

//MOTD
	PackPointer = PlayerList;

	NumConnected = 0;
	for (int hh=0;hh<128;hh++)
	{
		if (ConnUser[hh].Connected)
			NumConnected++;
	}

	sprintf(MOTDa, "Hello, %s, welcome to %s!\nThis server is running a Beta copy of ACServer.  Check www.acserver.org for updates.\n\nThere are currently: %i Clients Connected.\n\n", ConnUser[UserParsing].ZoneName, ServerName, NumConnected);

	DWORD Zero = 0x0000F65A;				memcpy(PackPointer,&Zero,4);	PackPointer += 4;		//F65A - DWORD
	GenString(MOTDa, &outlen, PackPointer);	PackPointer += outlen;
	GenString(MOTDb, &outlen, PackPointer); PackPointer += outlen;

	Send200(PlayerList, (int) (PackPointer - PlayerList), 4, UserParsing);
}

void cACServer::Send200(BYTE *data, int datalen, WORD wUnk, int User)
{
	EmuHeader tph;
	tph.payloadsize_ = datalen;
	tph.message_ = wUnk;
	send(ConnUser[User].Socket, (char *) &tph, sizeof(tph), NULL);

	send(ConnUser[User].Socket, (char *) data, datalen, NULL);

//	for (int i=0;i<datalen;i+=464)	//Split into 464 byte chunks 4 some reason... :P
//		send(ConnUser[User].Socket, (char *) data+i, (datalen - i >= 464) ? 464 : (datalen - i), NULL);
}

void cACServer::DisconnectUser(int User)
{
	if (ConnUser[User].Connected == false)
		return;

	ConnUser[User].ObjectCache.clear();

	ConnUser[User].Connected = false;

	closesocket(ConnUser[User].Socket);

	int NumConn = 0;
	for (int hh=0;hh<128;hh++)
	{
		if (ConnUser[hh].Connected)
			NumConn++;
	}

	ZonePasses[ConnUser[User].ZoneNum].LoggedIn = (WORD) -1;

	UpdateCharInfo(User, "Not Connected");

	//Recalc MaxUsers
	for (int i=127;i>=0;i--)
	{
		if (ConnUser[i].Connected)
		{	MaxUsers = i+1;	break; }
	}

	if (ConnUser[User].Char == 0)
		return;

	SaveCharacter(User);

	LBObjects[ConnUser[User].Char->Loc.landblock >> 16].erase(ConnUser[User].Char->GUID);

	ObjectList.erase(ConnUser[User].Char->GUID);

	//Return him to char select screen
	DWORD tph3 = 0x0000F653;

	Send200((BYTE *) &tph3, sizeof(tph3), 4, User);

	//Disconnect Him
	ConnUser[User].Connected = false;

	sprintf(sayit, "Player %s Disconnected!", ConnUser[User].Char->Name);
	ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);

	//Delete him from other people
	RemoveItem(ConnUser[User].Char->GUID);

	ConnUser[User].Char = 0;
}

void cACServer::UpdateCharInfo(int User, char * NewText)
{
	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_DELETESTRING, User+1, 0);

	char TpStr[100];
	strcpy(ConnUser[User].Status, NewText);
	if (ConnUser[User].Connected)
	{
		sprintf(TpStr, "%i\t%i.%i.%i.%i:%i\t%s", User,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b1,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b2,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b3,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b4,
			ntohs(ConnUser[User].sockaddy.sin_port),
			NewText);
	} else {
		sprintf(TpStr, "%i\tN/A\t%s", User, NewText);
	}
	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_INSERTSTRING, User+1, (WPARAM) &TpStr[0]);
}

void cACServer::ServerMessage(int User, char *Text, DWORD Color)
{
	try {
		BYTE TpMes[1024]; BYTE * PackPtr = &TpMes[0];
		DWORD tpf;
		int outlen;
		tpf = 0x0000F62C;		memcpy(PackPtr,&tpf,4);	PackPtr += 4;
		GenString(Text, &outlen, PackPtr); PackPtr += outlen;
		tpf = Color;		memcpy(PackPtr,&tpf,4);	PackPtr += 4;
		if ((User >= 0) && (User < 128))
		{
			Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, User);
		} else if (User == GLOBAL_MESSAGE) {
			//Global Message...
			for (int i=0;i<MaxUsers;i++)
			{
				if ((ConnUser[i].Connected) && (ConnUser[i].State >= 5))
					Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, i);
			}
		} else if ((User > -128) && (User < 0)) {
			//Global Message except User
			for (int i=0;i<MaxUsers;i++)
			{
				if ((ConnUser[i].Connected) && (ConnUser[i].State >= 5) && (i != (User*-1)-1))
					Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, i);
			}
		}
	} catch (char * ) {
		//Meh...
		
	}
}

void cACServer::SaveCharacter(int CharNum)
{
	FILE *charread = fopen("CharList.dat","r+b");

	fseek(charread, 0, SEEK_END);
	int numchars = ftell(charread)/(SIZEOFCHAR+68);
	fseek(charread, 0, SEEK_SET);

	DWORD GUIDtemp;
	for (int j=0;j<numchars;j++)
	{
		fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);
		fread(&GUIDtemp, 4, 1, charread);
		if (GUIDtemp == ConnUser[CharNum].Char->GUID)
		{
			fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);

			BYTE TPlog[5000]; int charsize = ConnUser[CharNum].Char->Serialize(TPlog);
			fwrite(&ConnUser[CharNum].Char->GUID, 4, 1, charread);
			fwrite(ConnUser[CharNum].Char->Name, 32, 1, charread);
			fwrite(&ConnUser[CharNum].Char->Loc, sizeof(Location_t), 1, charread);
			fwrite(TPlog, charsize, 1, charread);
			
//			fwrite(ConnUser[CharNum].Char, sizeof(cWorldObject_Char), 1, charread);
			j = numchars;
		}
	}

	fclose(charread);
}

int cACServer::Generate_F7B0_13(BYTE * OutPack, int User)
{
	BYTE *PackPointer = OutPack;
	DWORD tpd;	WORD tpw;
	
	//F7b0 header
	tpd = 0xF7B0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].Char->GUID, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].EventCount, 4);	PackPointer += 4;
	tpd = 0x0013;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//13 = Login Char
	ConnUser[User].EventCount++;
	
	//On the the actual packet... - loginmask1 - 7B in the stuff...
	tpd = 0x001B;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginMask1
	tpd = 0x000A;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown1 - A
	
	//count1
	tpw = 0x000C;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count for next section
	tpw = 0x0040;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown for this - 40

	tpd = 0x0081;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x02F7;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x0084;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	//Burden, calc later..
	tpd = 0x0005;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//5 = burden
	tpd =      1;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//1 = ?

	tpd = 0x008B;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0037;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	//Pyreals, calc later
	tpd = 0x0014;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//14 = pyreals
	tpd =      0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//0
	//Total Xp
	tpd = 0x0015;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//15 = total xp
						memcpy(PackPointer, &ConnUser[User].Char->TotalXP, 4);	PackPointer += 4;
	//Unassigned XP
	tpd = 0x0016;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//16 = unassigned xp
						memcpy(PackPointer, &ConnUser[User].Char->UnassignedXP, 4);	PackPointer += 4;
	//Unassigned Skill Points
	tpd = 0x0018;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//18 = unassigned skill points
						memcpy(PackPointer, &ConnUser[User].Char->SkillCredits, 4);	PackPointer += 4;
	//Level, calc later....
	tpd = 0x0019;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//19 = Level
						memcpy(PackPointer, &ConnUser[User].Char->Level, 4);	PackPointer += 4;
	//Rank, calc later
	tpd = 0x001e;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//1e = Rank
	tpd = 0x0000;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x0028;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x002F;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0000;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
//end of that vector

	//count2	
	tpw = 0x0005;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count2
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown - 20
	//count2 vector here... i.e. nothing yet

	//Count2 Stuff...
	DWORD CHAR_DATA_VECTOR2[] = {
		0x00000004, 0x00000000, 0x0000002C, 0x00000000, 
		0x0000002D, 0x00000000, 0x0000002E, 0x00000000, 
		0x0000002F, 0x00000000, 
	};
	memcpy(PackPointer, CHAR_DATA_VECTOR2, sizeof(CHAR_DATA_VECTOR2));	PackPointer += sizeof(CHAR_DATA_VECTOR2);

	//strings
	tpw = 0x0004;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//stringcount
	tpw = 0x0010;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//strings unknown - 10

	//Char Name
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						GenString(ConnUser[User].Char->Name, (int *) &tpd, PackPointer);	PackPointer += tpd;
	//Gender
	tpd = 0x0003;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	char Gender[7];
	if (ConnUser[User].Char->Gender == 0)
		sprintf(Gender, "Female");
//	if (ConnUser[User].Char->Gender == 1)
	else
		sprintf(Gender, "Male");
	GenString(Gender, (int *) &tpd, PackPointer);	PackPointer += tpd;

	//Race
	char Race[12];
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	if (ConnUser[User].Char->Race == 0)
		sprintf(Race, "Aluvian");
	if (ConnUser[User].Char->Race == 1)
		sprintf(Race, "Gharun'Dim");
//	if (ConnUser[User].Char->Race == 2)
	else
		sprintf(Race, "Sho");
	GenString(Race, (int *) &tpd, PackPointer);	PackPointer += tpd;

	//Class	- "Server Tester" for now... ;-)
	tpd = 0x0005;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						GenString("Server Tester", (int *) &tpd, PackPointer);	PackPointer += tpd;

	//count4
	tpw = 0x0001;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count4
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count4unknown - 20
	//count4 vector here... i.e. nothing yet

	//sec4 stuff
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//
	tpd = 0x30000000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//

	//Loginmask1 info goes here

	

	//End loginmask1 shit
	//should be 103, 100 = spellbook, 2 = skills
	tpd = 0x0103;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Loginmask2
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown14 - always 1
	tpd = 0x01ff;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//attributemask - always 0x1ff
	
	//Primary Stats						
	for (int i=0;i<6;i++)
	{
		if (i == 3)
		{
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[2], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[2], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[2], 4);	PackPointer += 4;
		} else if (i == 2) {
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[3], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[3], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[3], 4);	PackPointer += 4;
		} else {
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[i], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[i], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[i], 4);	PackPointer += 4;
		}
	}

	//Secondary Stats
	for (i=0;i<3;i++)
	{
		memcpy(PackPointer, &ConnUser[User].Char->SecondaryStatInc[i], 4);		PackPointer += 4;
		tpd = 0;	memcpy(PackPointer, &tpd, 4);		PackPointer += 4;	//?... always 0 afawct
		memcpy(PackPointer, &ConnUser[User].Char->XPIntoSecondaryStat[i], 4);		PackPointer += 4;
		memcpy(PackPointer, &ConnUser[User].Char->CurrentSecondaryStat[i], 4);		PackPointer += 4;
	}

	//Skillcount
	tpw = 0x0023;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown - 20
	for (i=1;i<40;i++)
	{
		if (ConnUser[User].Char->SkillTrain[i] > 0)
		{
			memcpy(PackPointer, &i, 4);	PackPointer += 4;		//Skill number
			memcpy(PackPointer, &ConnUser[User].Char->SkillInc[i], 4);	PackPointer += 4;		//Skill number
			memcpy(PackPointer, &ConnUser[User].Char->SkillTrain[i], 4);	PackPointer += 4;		//Trained Status
			memcpy(PackPointer, &ConnUser[User].Char->SkillXP[i], 4);	PackPointer += 4;		//Total XP into skill
			memcpy(PackPointer, &ConnUser[User].Char->SkillFreeXP[i], 4);	PackPointer += 4;	//Free XP at gen
			tpd = 0x0000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;	//unknown2
			
			//double thingy
			tpd = 0xE5C9D958;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
			tpd = 0x4171BA45;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
		}
	}

	//Loginmask2 time
	//100 - spellbook
	WORD numspells = 0;
	for (i=0;i<96;i++)
	{
		for (DWORD tpdd=ConnUser[User].Char->SpellsLearned[i];tpdd > 0;tpdd >>= 1)
		{
			if (tpdd & 1)
				numspells++;
		}
	}
						memcpy(PackPointer, &numspells, 2);	PackPointer += 2;		//count
	tpw = 0x0040;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//spellcountunknown

	int tpinc;	float tpcharge = 1.0f;
	for (i=0;i<96;i++)
	{
		tpinc = 0;
		for (DWORD TpSpell = ConnUser[User].Char->SpellsLearned[i]; TpSpell > 0; TpSpell >>= 1)
		{
			if (TpSpell & 1)
			{
				tpd = (i << 5) | tpinc;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//spell ID
											memcpy(PackPointer, &tpcharge, 4);	PackPointer += 4;		//charge
				tpd = 0;					memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown2
				tpd = 0;					memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown3
			}
			tpinc++;
		}
	}

	//200 - enchantments... whoooa there cowboy

	DWORD Numshorts = 0;
	for (i=0;i<10;i++)
		if (ConnUser[User].Char->ShortCut[i])
			Numshorts++;

	if (Numshorts)
		tpd = 0x0005;
	else
		tpd = 0x0004;
						memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3

	//4, character options
//	tpd = 0x01ECE563;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3
	tpd = 0x01E4E56A;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3

	if (Numshorts)
	{
		//1, shortcut bar
		memcpy(PackPointer, &Numshorts, 4);	PackPointer += 4;
		for (i=0;i<10;i++)
			if (ConnUser[User].Char->ShortCut[i])
			{
				memcpy(PackPointer, &i, 4);	PackPointer += 4;									//Position
				memcpy(PackPointer, &ConnUser[User].Char->ShortCut[i], 4);	PackPointer += 4;	//Object
				tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;					//Unknown - 0
			}
	}

	//Spellbars
	for (i=0;i<5;i++)
	{
		DWORD numspells = 0;
		for (numspells=0;ConnUser[User].Char->SpellBar[i][numspells] > 0;numspells++) { }

		memcpy(PackPointer, &numspells, 4);	PackPointer += 4;
		for (int j=0;j<numspells;j++)
		{
			memcpy(PackPointer, &ConnUser[User].Char->SpellBar[i][j], 4);	PackPointer += 4;
		}
	}

	//comp buyer...
	
	//Inventory
	DWORD numinv = 0;
	for (i=0;i<7;i++)
		if (ConnUser[User].Char->SidePacks[i])
			numinv++;
	for (i=0;i<102;i++)
		if (ConnUser[User].Char->MainPack[i])
			numinv++;

	memcpy(PackPointer, &numinv, 4);	PackPointer += 4;
	for (i=0;i<7;i++)
		if (ConnUser[User].Char->SidePacks[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->SidePacks[i], 4);	PackPointer += 4;
			tpd = 1;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//IsContainer = 1
		}
	for (i=0;i<102;i++)
		if (ConnUser[User].Char->MainPack[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->MainPack[i], 4);	PackPointer += 4;
			tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//IsContainer = 0
		}

	//Equipped Shizzit
	DWORD numequip = 0;
	for (i=0;i<25;i++)
		if (ConnUser[User].Char->Equipped[i])
			numequip++;

	memcpy(PackPointer, &numequip, 4);	PackPointer += 4;
	for (i=0;i<25;i++)
		if (ConnUser[User].Char->Equipped[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->Equipped[i], 4);	PackPointer += 4;
			tpd = (1 << i);	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Coverage
			tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown3
		}

	return((int) (PackPointer - OutPack));
}

void cACServer::LoadObjects()
{
	sprintf(sayit, "Loading Objects.Dat..."); LogDisp(sayit);

	FILE *in = fopen("Objects.dat","rb");
	if (!in)
	{
		sprintf(sayit, "Objects.Dat Empty!  Initializing with an empty empty world."); LogDisp(sayit);
		return;
	}

	BYTE ObjData[4096];
	cWorldObject *tpObj = (cWorldObject *) ObjData;
	
	int errtest;

	cWorldObject *tpo;

	while (!feof(in))
	{
		DWORD ObjType;  errtest = fread(&ObjType, 4, 1, in);
	
		if (!errtest)
		{
			fclose(in);
			sprintf(sayit, "Objects.Dat Loaded!"); LogDisp(sayit);
			return;
		}

		switch (ObjType)
		{
		case OBJECT_LIFESTONE:
			tpo = new cWorldObject_LS();
			break;
		case OBJECT_OBJECTID:
			tpo = new cWorldObject_ObjectID();
			break;
		case OBJECT_PORTAL:
			tpo = new cWorldObject_Portal();
			break;
		case OBJECT_CHARACTER:
			//Ignore
			MessageBox(NULL, "Fatal Error: Character found in objects.dat...", "Fatal Error", MB_OK);
			break;
		default:
			//eep... bad.
			MessageBox(NULL, "Fatal Error: Unknown Object type in Objects.Dat.  Please Restore Objects.Dat.  Objects.Dat loading halted...", "Fatal Error", MB_OK);
			return;
		}

		if (ObjType != OBJECT_CHARACTER)
		{
			fread(&tpo->GUID, 4, 1, in);
			fread(tpo->Name, 32, 1, in);
			fread(&tpo->Loc, 32, 1, in);

			fread(ObjData, tpo->Size, 1, in);
			tpo->LoadFrom(ObjData);
			ObjectList[tpo->GUID] = tpo;
			LBObjects[tpo->Loc.landblock >> 16][tpo->GUID] = tpo;
		}

		if (ObjType == OBJECT_PORTAL)
			Portals[tpo->Loc.landblock >> 16][tpo->GUID] = (cWorldObject_Portal *) tpo;
	}
}

void cACServer::SaveObjects()
{
//	DeleteFile("Objects.Bak");
//	rename("Objects.Dat", "Objects.Bak");

	FILE *out = fopen("Objects.Dat","wb");
	if (!out)
	{
		//bad...
		MessageBox(NULL, "Can't open objects.dat for writing... very bad...", "Fatal Error", MB_OK);
		return;
	}
	
	BYTE SerData[4096];	int SerSize;

	for (std::map<DWORD, cWorldObject *>::iterator i = ObjectList.begin(); i != ObjectList.end(); i++)
	{
		if ((i->second) && (i->second->Type != OBJECT_CHARACTER))
		{
			fwrite(&i->second->Type, 4, 1, out);
			fwrite(&i->second->GUID, 4, 1, out);
			fwrite(i->second->Name, 32, 1, out);
			fwrite(&i->second->Loc, 32, 1, out);

			SerSize = i->second->Serialize(SerData);
			if (SerSize)
				fwrite(SerData, SerSize, 1, out);
		}
	}

	fclose(out);
}

void cACServer::SendLandblock(int User, WORD Landblock)
{
	int packlen;
	BYTE TpP[500];
	
	for (std::map<DWORD, cWorldObject *>::iterator i = LBObjects[Landblock].begin(); i != LBObjects[Landblock].end(); i++)
	{
		if ((i->second->Type != 0) && (i->second->GUID != ConnUser[User].Char->GUID))
		{
			packlen = i->second->GenerateCreatePacket(TpP);
			Send200(TpP, packlen, 3, User);
		}
	}
}

void cACServer::MoveUser(int User, Location_t *loc, bool Portal, bool Override)
{
	BYTE MOVE_ENTITY[] = {
		0x48, 0xF7, 0x00, 0x00, 0x03, 0x40, 0x16, 0x78, 0x34, 0x00, 0x00, 0x00, 0x08, 0x01, 0x64, 0x81,   //(H___@_x4_____d_) - 0000
		0xA1, 0xDB, 0x18, 0x41, 0xF0, 0x07, 0x19, 0x43, 0xD7, 0xA3, 0xBD, 0x41, 0xAA, 0x77, 0x6D, 0x3F,   //(___A___C___A_wm?) - 0010
		0xE0, 0x42, 0xBF, 0xBE, 0x39, 0x06, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,                           //(_B__9_______)     - 0020
	};

	for (std::map<DWORD, cWorldObject *>::iterator i = Portals[loc->landblock >> 16].begin(); i != Portals[loc->landblock >> 16].end(); i++)
	{
		if (GetDistance(loc, &i->second->Loc) < 0.01)
		{
			MoveUser(User, &((cWorldObject_Portal *)i->second)->DestLoc, true, false);
//			ServerMessage(User, "Entered Portal... Teleporting...", COLOR_RED);
		}
	}

	float heading = acos(loc->a);
	if (loc->w > 0) heading = PI - heading;
	heading *= 2;
	ConnUser[User].PointX = sin(heading);
	ConnUser[User].PointY = cos(heading);

	if (!ConnUser[User].Char)
		return;

	WORD OldLB = ConnUser[User].Char->Loc.landblock >> 16, NewLB = loc->landblock >> 16;
	
	memcpy(&ConnUser[User].Char->Loc, loc, sizeof(Location_t));

	DWORD tpf2 = 0x0000F748; memcpy (&MOVE_ENTITY[0x00],&tpf2,4);
	memcpy(&MOVE_ENTITY[0x4], &ConnUser[User].Char->GUID, 4);
	tpf2 = 0x00000034; memcpy (&MOVE_ENTITY[0x08],&tpf2,4);
	memcpy(&MOVE_ENTITY[0x0C], &ConnUser[User].Char->Loc, 4*5);
	memcpy(&MOVE_ENTITY[0x20], &ConnUser[User].Char->Loc.w, 4);

	if (Portal)
		ConnUser[User].PortalCount++;
	if (Override)
		ConnUser[User].OverrideCount++;
	
	memcpy (&MOVE_ENTITY[0x24],&ConnUser[User].Char->NumLogins,2);
	memcpy (&MOVE_ENTITY[0x26],&ConnUser[User].move_count,2);
	memcpy (&MOVE_ENTITY[0x28],&ConnUser[User].PortalCount,2);
	memcpy (&MOVE_ENTITY[0x2A],&ConnUser[User].OverrideCount,2);

	for (int iii=0;iii<MaxUsers;iii++)
	{
		if ((ConnUser[iii].Connected) && ((ConnUser[iii].State == 6) || (iii == User)))
			Send200(&MOVE_ENTITY[0], sizeof(MOVE_ENTITY), 3, iii);
	}
	ConnUser[User].move_count++;

	if (OldLB != NewLB)
	{
		LBObjects[OldLB].erase(ConnUser[User].Char->GUID);
		LBObjects[NewLB][ConnUser[User].Char->GUID] = (cWorldObject *) ConnUser[User].Char;


		if (OldLB == NewLB - 1) {
			SendLandblock(User, NewLB - 1);
			SendLandblock(User, NewLB - 1 - 256);
			SendLandblock(User, NewLB - 1 + 256);
		} else if (OldLB == NewLB + 1) {
			SendLandblock(User, NewLB + 1);
			SendLandblock(User, NewLB + 1 - 256);
			SendLandblock(User, NewLB + 1 + 256);
		} else if (OldLB == NewLB - 256) {
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256);
			SendLandblock(User, NewLB - 256 + 1);
		} else if (OldLB == NewLB + 256) {
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256);
			SendLandblock(User, NewLB + 256 + 1);
//Diagonals Now
		} else if (OldLB == NewLB + 256 + 1) {
			SendLandblock(User, NewLB + 256 + 1);
			SendLandblock(User, NewLB + 256 + 1 - 256);
			SendLandblock(User, NewLB + 256 + 1 - 256 - 256 );
			SendLandblock(User, NewLB + 256 + 1 - 1);
			SendLandblock(User, NewLB + 256 + 1 - 1 - 1);
		} else if (OldLB == NewLB + 256 - 1) {
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256 - 1 - 256);
			SendLandblock(User, NewLB + 256 - 1 - 256 - 256 );
			SendLandblock(User, NewLB + 256 - 1 + 1);
			SendLandblock(User, NewLB + 256 - 1 + 1 + 1);
		} else if (OldLB == NewLB - 256 + 1) {
			SendLandblock(User, NewLB - 256 + 1);
			SendLandblock(User, NewLB - 256 + 1 + 256);
			SendLandblock(User, NewLB - 256 + 1 + 256 + 256 );
			SendLandblock(User, NewLB - 256 + 1 - 1);
			SendLandblock(User, NewLB - 256 + 1 - 1 - 1);
		} else if (OldLB == NewLB - 256 - 1) {
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256 - 1 + 256);
			SendLandblock(User, NewLB - 256 - 1 + 256 + 256 );
			SendLandblock(User, NewLB - 256 - 1 + 1);
			SendLandblock(User, NewLB - 256 - 1 + 1 + 1);
		} else {
			SendLandblock(User, NewLB);
			SendLandblock(User, NewLB - 1);
			SendLandblock(User, NewLB + 1);
			SendLandblock(User, NewLB - 256);
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256 + 1);
			SendLandblock(User, NewLB + 256);
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256 + 1);
		}
	}
}

void cACServer::SendAnim(int User, WORD AnimNum)
{
	ConnUser[User].AnimCount++;

	cAnimStruct ASt;
	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0;
	ASt.wAnimation_Family = 0x3D;
	ASt.wUnk2 = 0x80;
	ASt.wUnk3 = 0;
	ASt.wAnimation_Seqnum = ConnUser[User].AnimCount;
	ASt.wAnimation_to_Play = (WORD) AnimNum;
	ASt.fPlaySpeed = 1.0f;

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ASt, sizeof(ASt), 3, i);
	}
}

void cACServer::PointAt(int User, DWORD GUIDAt)
{
	ConnUser[User].AnimCount++;

	cAnimStructEx ASt;
	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0;
	ASt.wAnimation_Family = 0x3D;

	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0x0008;
	ASt.wAnimation_Family = 0x3D;
	ASt.GUIDTarget = GUIDAt;
	ASt.dwUnk1 = 0x429DB3C7;
	ASt.dwUnk2 = 0x0155EE0F;
	ASt.fUnk1 = 1.0f;
	ASt.Zero = 0;

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ASt, sizeof(ASt), 3, i);
	}

}

void cACServer::SendParticle(int User, DWORD ParticleNum)
{
	DWORD ParticleEffect[] = {
		0xF755,
		ConnUser[User].Char->GUID,
		ParticleNum,
		0
	};

	float tpf = 1.0f;
	memcpy(&ParticleEffect[3],&tpf,4);

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ParticleEffect, sizeof(ParticleEffect), 3, i);
	}
}

void cACServer::GiveXP(int User, int Type, DWORD Amount)
{
	BYTE Pack237[0xD];
	DWORD tpdword;	BYTE tpbyte;

	if ((User > 128) || (!ConnUser[User].Connected) || (ConnUser[User].State != 6))
	{
		ServerMessage(User, "Error in !givexp format... Make sure you're giving a valid User #...", COLOR_RED);
		
		return;
	}

	ConnUser[User].Char->TotalXP += Amount;

	ZeroMemory(Pack237, 0xD);
	tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
	tpbyte = ConnUser[User].Count237[0x15];	memcpy(&Pack237[4], &tpbyte, 1);
	tpdword = 0x15;	 memcpy(&Pack237[5], &tpdword, 4);		//Total XP
	tpdword = ConnUser[User].Char->TotalXP;  memcpy(&Pack237[9], &tpdword, 4);
	Send200(Pack237, sizeof(Pack237), 4, User);
	ConnUser[User].Count237[0x15]++;

	int NewLevel = 0;
	for (int i=ConnUser[User].Char->Level;i<128;i++)
	{
		if (ConnUser[User].Char->TotalXP < LevelArray[i])
		{
			NewLevel = i - 1;

			i = 128;
		}
	}
	if (NewLevel > ConnUser[User].Char->Level)
	{
		//Level him!
		SendParticle(User, 0x89);
		
		//Give skill credits as appropriate...
		for (i=ConnUser[User].Char->Level+1;i<=NewLevel;i++)
		{
			for (int h=0;h<=sizeof(CreditLevel)/4;h++)
				if (CreditLevel[h] == i)
					ConnUser[User].Char->SkillCredits++;
		}
		
		ConnUser[User].Char->Level = NewLevel;
		
		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x19];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x19;	 memcpy(&Pack237[5], &tpdword, 4);		//Level
		tpdword = ConnUser[User].Char->Level;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x19]++;

		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x18];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x18;	 memcpy(&Pack237[5], &tpdword, 4);		//Credits
		tpdword = ConnUser[User].Char->SkillCredits;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x18]++;

		sprintf(sayit, "You are now level %i!", ConnUser[User].Char->Level);
		ServerMessage(User, sayit, COLOR_CYAN);
		char mults[2] = { 's', 0 };
		if (ConnUser[User].Char->SkillCredits == 1)
			mults[0] = 0;
		printf(sayit, "You have %u experience points and %i skill credit%s available to raise skills and attributes.", ConnUser[User].Char->UnassignedXP, ConnUser[User].Char->SkillCredits, mults);
		ServerMessage(User, sayit, COLOR_CYAN);
		int NextCredit = 0, i=0;
		while (!NextCredit)
		{
			if (ConnUser[User].Char->Level < CreditLevel[i])
				NextCredit = CreditLevel[i];
			
			i++;
		}
		sprintf(sayit, "You will earn another skill credit at level %i.", NextCredit);
		ServerMessage(User, sayit, COLOR_CYAN);
	}

	if ((Type < 0) || ((Type > 39) && (Type != 128)))
	{
		ServerMessage(User, "Error in !givexp format... Type should be from 0-39 or 128 for unassigned XP...", COLOR_RED);

		return;
	}

	if (Type == SKILL_UNASSIGNED)
	{
		ConnUser[User].Char->UnassignedXP += Amount;
		
		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x16];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x16;	 memcpy(&Pack237[5], &tpdword, 4);		//Unassigned XP
		tpdword = ConnUser[User].Char->UnassignedXP;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x16]++;
	} else if (Type < 40) {
		ConnUser[User].Char->SkillXP[Type] += Amount;

		int NewSkillLevel = 0;
		for (int i=0;i<200;i++)
		{
			if (ConnUser[User].Char->SkillXP[Type] < SkillArray[i])
			{
				NewSkillLevel = i - 1;

				i = 200;
			}
		}
		if (NewSkillLevel > ConnUser[User].Char->SkillInc[Type])
		{
			ConnUser[User].Char->SkillInc[Type] = NewSkillLevel;
		}

		//Send skill xp packet
	}
}

void cACServer::ActionComplete(int User)
{
	DWORD ActComp[] = {
		0x0000F7B0,
		ConnUser[User].Char->GUID,
		ConnUser[User].EventCount,
		0x000001C7,
		0x00000000
	};
	ConnUser[User].EventCount++;
	Send200((BYTE *) ActComp, sizeof(ActComp), 3, User);
}

float GetDistance(Location_t *loc1, Location_t *loc2)
{
	DWORD landblock;
	int XX, ns, ew;
	double xc, yc, PosNSa, PosEWa, PosNSb, PosEWb;

	float Distance;

	if ((loc1->landblock & 0x0000FF00) && (loc2->landblock & 0x0000FF00))
	{
		//Interior
		Distance = fabs(sqrt(pow(loc2->x - loc1->x,2) + pow(loc2->y - loc1->y,2) +  + pow(loc2->z - loc1->z,2)))/240;
	} else {
		//Exterior
		landblock = loc1->landblock;
		xc = loc1->x;
		yc = loc1->y;

		XX = landblock & 0x3F; 
		ns = ((((landblock & 0x00FF0000) >> 16) + 1) * 8) + ((XX - 1) & 7) - 1027;
		ew = ((((landblock & 0xFF000000) >> 24) + 1) * 8) + (int) ((XX - 1) / 8) - 1027;
    
		xc = xc - ((int) (xc / 24.0f) * 24.0f);
		yc = yc - ((int) (yc / 24.0f) * 24.0f);
		PosEWa = ((float) (ew) / 10.0f) + (xc / 240.0f);
		PosNSa = ((float) (ns) / 10.0f) + (yc / 240.0f);
		
		landblock = loc2->landblock;
		xc = loc2->x;
		yc = loc2->y;

		XX = landblock & 0x3F; 
		ns = ((((landblock & 0x00FF0000) >> 16) + 1) * 8) + ((XX - 1) & 7) - 1027;
		ew = ((((landblock & 0xFF000000) >> 24) + 1) * 8) + (int) ((XX - 1) / 8) - 1027;

		xc = xc - (float) ((int) (xc / 24.0f) * 24.0f);
		yc = yc - (float) ((int) (yc / 24.0f) * 24.0f);
		PosEWb = ((float) (ew) / 10.0f) + (xc / 240.0f);
		PosNSb = ((float) (ns) / 10.0f) + (yc / 240.0f);

		Distance = fabs(sqrt(pow(PosNSa - PosNSb,2) + pow(PosEWa - PosEWb,2) + pow((loc1->z - loc2->z)/240,2)));
	}

	return(Distance);
}

void cACServer::WorldServerParse(int UserParsing, BYTE *tPackPtr, int amountRead, cClientPacketHeader *tphdr)
{
	if (tphdr->m_dwFlags == 0x00000200) {
		DWORD MType, MType2;
		BYTE * tPackPtr;
		tPackPtr = &TestBuf[sizeof(cClientPacketHeader)];
		while (tPackPtr < &TestBuf[amountRead])
		{
			if (!ConnUser[UserParsing].Char)
				return;
			
			memcpy(&MType,tPackPtr,4);
			memcpy(&MType2,&tPackPtr[8],4);

			if (MType == 0x0000F653) {
				//Exit world..
//				DisconnectUser(UserParsing);

				SaveCharacter(UserParsing);

				LBObjects[ConnUser[UserParsing].Char->Loc.landblock >> 16].erase(ConnUser[UserParsing].Char->GUID);

				ObjectList.erase(ConnUser[UserParsing].Char->GUID);

				//Return him to char select screen
				DWORD tph3 = 0x0000F653;

				Send200((BYTE *) &tph3, sizeof(tph3), 4, UserParsing);

				//Disconnect Him
				sprintf(sayit, "Player %s Disconnected!", ConnUser[UserParsing].Char->Name);
				ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);

				//Delete him from other people
				//Somehow...

				ConnUser[UserParsing].Char = 0;

				ConnUser[UserParsing].State = 1;

				SendCharList(UserParsing);
			} else if (MType == 0xF6EA) {
				//Nobody knows...
			} else if (MType == 0xF7A9) {
				//Client needs landblock
				
				ServerMessage(UserParsing, "* You don't have at least one of the landblocks in your new area.  !telemap to another location or try typing /render radius 5.", COLOR_RED);
			} else if (MType == 0xF7B1) {
				if (MType2 == 0xF753)
				{
					//High-Priority Movement Packet

					Location_t *loc = (Location_t *) &tPackPtr[12];

					MoveUser(UserParsing, loc, false, false);
				} else if (MType2 == 0xF61C) {
					//Low-Priority Movement Packet

					DWORD *hex=(DWORD *)tPackPtr;

					Location_t *loc;	DWORD skip = 0;
					DWORD hex3;
					for (hex3=hex[3];hex3>0;hex3>>=1)
					{	if (hex3 & 1) skip++;	}

					//Flags
					DWORD SpeedFlag = 0, TurnFlag = 0, SlideFlag = 0, MoveFlag = 0, EmoteFlag = 0;

					DWORD * hexat = &hex[4];
					cAnimStruct ASt;
					ASt.wAnimation_Family = 0x3D;
					ASt.wAnimation_to_Play = 0x0;
					ASt.fPlaySpeed = 1.0f;
					if (hex[3] & 0x00000001)
					{	//Normal Speed, Unset = Walking
						SpeedFlag = *hexat;
						hexat += 1; }
					if (hex[3] & 0x00000004)
					{	//Moving
						DWORD MoveFlag = *hexat;
						hexat += 1;

						if (MoveFlag == 0x45000005)
						{
							ASt.fPlaySpeed = 1.0f;
							if (SpeedFlag == 0)
								ASt.wAnimation_to_Play = 0x05;
							else if (SpeedFlag == 2)
								ASt.wAnimation_to_Play = 0x07;
						} else if (MoveFlag == 0x45000006) {
							ASt.fPlaySpeed = -1.0f;
							ASt.wAnimation_to_Play = 0x05;
						} else if ((MoveFlag & 0xFF000000) == 0x41000000) {
							ASt.wAnimation_to_Play = (WORD) (MoveFlag & 0xFF);
						} else if ((MoveFlag & 0xFF000000) == 0x43000000) {
							ASt.wAnimation_to_Play = (WORD) (MoveFlag & 0xFF);
						} else {
							sprintf(sayit, "MoveFlag: %08X", MoveFlag); LogDisp(sayit);
						}
					}
					if (hex[3] & 0x00000020)
					{	//Sliding
						DWORD SlideFlag = *hexat;
						hexat += 1;

						if (SlideFlag == 0x6500000F)
						{
							ASt.fPlaySpeed = 1.0f;
							ASt.wAnimation_to_Play = 0x0F;
						} else if (SlideFlag == 0x65000010) {
							ASt.fPlaySpeed = -1.0f;
							ASt.wAnimation_to_Play = 0x0F;
						}
					}
					if (hex[3] & 0x00000100)
					{	//Turning
						DWORD TurnFlag = *hexat;
						hexat += 1;
						
						if (TurnFlag == 0x6500000D)			//Right
						{
							ASt.fPlaySpeed = 1.0f;
							ASt.wAnimation_to_Play = 0x0D;
						} else if (TurnFlag == 0x6500000E) {	//Left
							ASt.fPlaySpeed = -1.0f;
							ASt.wAnimation_to_Play = 0x0D;
						}
					}
					if (hex[3] & 0x0000F800)
					{	//Emote Backlog
						DWORD EmoteFlag = *hexat;
						DWORD numEmotes = (hex[3] & 0x0000F800) >> 11;
						hexat += (numEmotes * 2);
						skip += (numEmotes * 2) - 1;
						if (numEmotes > 2)	//For some reason, on 3+ emotes, it has another DWORD at the end
							skip++;
						
						//Just play the first one
						ASt.wAnimation_to_Play = (WORD) (EmoteFlag & 0xFF);
						if (numEmotes > 1)
						{
							//Lots of emotes...  Learn how to chain later! =]
						}
					}

					loc=(Location_t *)&hex[4+skip];
					MoveUser(UserParsing, loc, false, false);

					//Make Anim Packet... Ho ho ho!
					ASt.dwf74c = 0x0000F74C;
					ASt.dwGUID = ConnUser[UserParsing].Char->GUID;
					ASt.wNumInteractions = ConnUser[UserParsing].move_count;
					ASt.wNumLogins = ConnUser[UserParsing].Char->NumLogins;

					ConnUser[UserParsing].AnimCount++;
					ASt.wNumAnimations = ConnUser[UserParsing].AnimCount;
					ASt.wAnimation_Seqnum = ConnUser[UserParsing].AnimCount;

					ASt.wFlag = 0;
					ASt.wUnk1 = 0;
					ASt.GUIDTarget = 0x00000080;

					for (int iii=0;iii<MaxUsers;iii++)
					{
						if ((ConnUser[iii].Connected) && (ConnUser[iii].State == 6) && (iii != UserParsing))
							Send200((BYTE *) &ASt, sizeof(ASt), 3, iii);
					}

					cAnimStruct AStA;
					memcpy(&AStA, &ASt, sizeof(AStA));
					AStA.wFlag = 1;
					AStA.fPlaySpeed = ASt.fPlaySpeed;
					Send200((BYTE *) &AStA, sizeof(AStA), 3, UserParsing);

					ConnUser[UserParsing].move_count++;
				} else if (MType2 == 0x08) {
					//Player Attacked
					DWORD PlayerAttacked; memcpy(&PlayerAttacked, &tPackPtr[12], 4);
					DWORD AttackSpot; memcpy(&AttackSpot, &tPackPtr[16], 4);
					DWORD Slider; memcpy(&Slider, &tPackPtr[20], 4);
					
					//Spot: Low/mid/high, 3/2/1

					cWorldObject_Char *tpc = (cWorldObject_Char *) ObjectList[PlayerAttacked];
					if (tpc)
					{
						//... bad...?
						return;
					}
					DWORD TotalSkill = ConnUser[UserParsing].Char->SkillInc[0x0D] + (ConnUser[UserParsing].Char->CurrentStat[0] + ConnUser[UserParsing].Char->CurrentStat[2])/3;
					DWORD DamageDone = (rand() & 3) + (TotalSkill / 10) + (ConnUser[UserParsing].Char->CurrentStat[0]/20);
					
					tpc->CurrentSecondaryStat[0] -= DamageDone;

					//Damage Packet

					//Recieve
					BYTE DamagePacket[100], *PackPointer = DamagePacket;

					DWORD tpd = 0xF7B0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
										memcpy(PackPointer, &tpc->GUID, 4);	PackPointer += 4;
										memcpy(PackPointer, &ConnUser[tpc->Owner].EventCount, 4);	PackPointer += 4;
					tpd = 0x01b2;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//13 = Login Char
					ConnUser[tpc->Owner].EventCount++;
					int strlen; GenString(ConnUser[UserParsing].Char->Name, &strlen, PackPointer);	PackPointer += strlen;

					DWORD damage = 4; memcpy(PackPointer, &damage, 4); PackPointer += 4;
					double severity = 0.5f; memcpy(PackPointer, &severity, 8); PackPointer += 8;
								memcpy(PackPointer, &DamageDone, 4); PackPointer += 4;
					Send200(DamagePacket, (DWORD) PackPointer - (DWORD) DamagePacket, 4, tpc->Owner);
					
					//Inflict
					PackPointer = DamagePacket;

					tpd = 0xF7B0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
										memcpy(PackPointer, &ConnUser[UserParsing].Char->GUID, 4);	PackPointer += 4;
										memcpy(PackPointer, &ConnUser[UserParsing].EventCount, 4);	PackPointer += 4;
					tpd = 0x01b1;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//13 = Login Char
					ConnUser[UserParsing].EventCount++;
					GenString(tpc->Name, &strlen, PackPointer);	PackPointer += strlen;

					memcpy(PackPointer, &damage, 4); PackPointer += 4;
					memcpy(PackPointer, &severity, 8); PackPointer += 8;
								memcpy(PackPointer, &DamageDone, 4); PackPointer += 4;
					DWORD spot;
					switch (AttackSpot)
					{
					case 1:
						//High
						spot = rand() & 1;
						break;
					case 2:
						//Mid
						spot = 1 + (rand() % 5);
						break;
					case 3:
						//Low
						spot = 7 + (rand() & 1);
						break;
					}
									memcpy(PackPointer, &spot, 4); PackPointer += 4;

					Send200(DamagePacket, (DWORD) PackPointer - (DWORD) DamagePacket - 4, 4, UserParsing);

					//Attack Completed
					DWORD tp1a7[] = {
						0x0000F7B0,
						ConnUser[UserParsing].Char->GUID,
						ConnUser[UserParsing].EventCount,
						0x01A7,
						ConnUser[UserParsing].AttackCount
					};
					Send200((BYTE *) tp1a7, sizeof(tp1a7), 4, UserParsing);
					ConnUser[UserParsing].AttackCount++;
					ConnUser[UserParsing].EventCount++;

					if ((int) tpc->CurrentSecondaryStat[0] <= 0)
					{
						//Dead ktnxunf
						sprintf(sayit, "%s was 0wn3d by %s!", tpc->Name, ConnUser[UserParsing].Char->Name);
						ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);

						tpc->CurrentSecondaryStat[0] = (tpc->CurrentStat[1]/2) + tpc->SecondaryStatInc[0];

						MoveUser(tpc->Owner, &tpc->LSTie, true, false);
					}
				} else if (MType2 == 0x45) {
					//Player raised stat
					//DWORD 1 = ?
					//DWORD 2 = ?
					DWORD dword1; memcpy(&dword1, &tPackPtr[12], 4);
					DWORD dword2; memcpy(&dword2, &tPackPtr[16], 4);
				} else if (MType2 == 0x47) {
					//Player trained skill
					//DWORD 1 = skill number
					//DWORD 2 = ?
					DWORD dword1; memcpy(&dword1, &tPackPtr[12], 4);
					DWORD dword2; memcpy(&dword2, &tPackPtr[16], 4);
					
					//subtract skill points
					ConnUser[UserParsing].Char->SkillCredits -= CreditCost[dword1];
					//raise skill to trained
					ConnUser[UserParsing].Char->SkillTrain[dword1] = 2;

					//Update client
					
					//Skill credit number
					BYTE Pack237[0xD];
					DWORD tpdword;	BYTE tpbyte;
					ZeroMemory(Pack237, 0xD);
					tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
					tpbyte = ConnUser[UserParsing].Count237[0x18];	memcpy(&Pack237[4], &tpbyte, 1);
					tpdword = 0x18;	 memcpy(&Pack237[5], &tpdword, 4);		//Credits
					tpdword = ConnUser[UserParsing].Char->SkillCredits;  memcpy(&Pack237[9], &tpdword, 4);
					Send200(Pack237, sizeof(Pack237), 4, UserParsing);
					ConnUser[UserParsing].Count237[0x18]++;

					//Skill update - send 23E
				} else if (MType2 == 0x36) {
					//Player used item
					DWORD tpguid; memcpy(&tpguid, &tPackPtr[12], 4);
					cWorldObject * tpitem = ObjectList[tpguid];
					switch (tpitem->Type)
					{
					case OBJECT_LIFESTONE:
						memcpy(&ConnUser[UserParsing].Char->LSTie, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));
						ServerMessage(UserParsing, "You successfully tie to the lifestone!", COLOR_BLUE);
						SendAnim(UserParsing, 0x0036);
						break;
					case OBJECT_PORTAL:
						PointAt(UserParsing, tpguid);
//									SendAnim(UserParsing, 7);	//Run
//									ActionComplete(UserParsing);
						break;
					}
				} else if (MType2 == 0x63) {
					//Player /lifestoned :)
					sprintf(sayit, "%s died!", ConnUser[UserParsing].Char->Name);
					ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);

//								SendAnim(UserParsing, 0x0008);
					
					MoveUser(UserParsing, &ConnUser[UserParsing].Char->LSTie, true, false);
				} else if (MType2 == 0xA1) {
					//Character logged in...
					
					UpdateCharInfo(UserParsing, "Fully Logged In");

					//Fully log him in on others
					DWORD F74B[] = {
						0x0000F74B,
						ConnUser[UserParsing].Char->GUID,
						0x00400410,
						0x00020003
					};
//					memcpy(&F74B[3], &ConnUser[UserParsing].Char->NumLogins, 2);
//					memcpy((&F74B[3]) + 2, &ConnUser[UserParsing].PortalCount, 2);

					for (int iii=0;iii<MaxUsers;iii++)
					{
						if ((ConnUser[iii].Connected) && (ConnUser[iii].State == 6))
							Send200((BYTE *) F74B, sizeof(F74B), 3, iii);
					}
					
					BYTE CreatePlayer[0x500];	int packlen;
					for (iii=0;iii<MaxUsers;iii++)
					{
						if ((ConnUser[iii].Connected) && (ConnUser[iii].State == 6) && (iii != UserParsing))
						{
							packlen = ConnUser[iii].Char->GenerateCreatePacket(CreatePlayer);
							//Show others on him
							Send200(CreatePlayer, packlen, 3, UserParsing);

							memcpy(&F74B[1], &ConnUser[iii].Char->GUID, 4);
							Send200((BYTE *) &F74B[0], sizeof(F74B), 3, UserParsing);
						}
					}
				} else if (MType2 == 0x00000015) {
					//Text Message from Client
					if ((int) (tphdr->m_wTotalSize-14) > 0)
					{
						char messagetext[2000];
						memcpy(messagetext,&tPackPtr[14],tphdr->m_wTotalSize-14);

						if (*messagetext == '!')
						{
							//Command
							char command[50], *Contents; ZeroMemory(command, 50);
//										char seps[] = " ";
							if (strstr(messagetext, " "))
							{
								memcpy(command,messagetext,(DWORD) strstr(messagetext, " ") - (DWORD) messagetext);
								Contents = strstr(messagetext," ");
							} else {
								memcpy(command,messagetext,strlen(messagetext));
								Contents = 0;
							}
							if (Contents)
								Contents = strstr(Contents, " ") + 1;
							
							int NumParms = 0;
							char Parms[10][20];	ZeroMemory(Parms, 10*20);
							if ((DWORD) Contents > 1)
							{
								char *tp = /*strstr(Contents," ") + 1*/ Contents, *tp2;
								while ((DWORD) tp > 1)
								{
									tp2 = strstr(tp," ") + 1;
									if ((DWORD) tp2 <= 1)
									{
										memcpy(Parms[NumParms], tp, strlen(tp));
										NumParms++;
										tp = tp2;
									} else {
										memcpy(Parms[NumParms], tp, (DWORD) (tp2 - tp - 1));
										NumParms++;
										tp = tp2;
									}
								}
							}

							if (stricmp(&command[1],"tele") == 0)	//Teleport by landblock
							{
								try {
									Location_t templ;
									memcpy(&templ, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));

									if (NumParms < 4)
									{
										ServerMessage(UserParsing, "Error in !tele syntax, check !help!", COLOR_RED);
										return;
									}
									//Fill in new Loc...
									sscanf(Parms[0],"%f",&templ.x);
									sscanf(Parms[1],"%f",&templ.y);
									sscanf(Parms[2],"%f",&templ.z);
									sscanf(Parms[3],"%08X",&templ.landblock);

									MoveUser(UserParsing, &templ, true, false);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !tele syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"telemap") == 0) {
								try {
									if (NumParms < 2)
									{
										ServerMessage(UserParsing, "Error in !telemap syntax, check !help!", COLOR_RED);
										return;
									}
									float NS, EW;//, Nx, Ny;
									char tpt;
									tpt = Parms[0][strlen(Parms[0])-1];
									Parms[0][strlen(Parms[0])-1] = 0;
									if ((tpt == 'N') || (tpt == 'n'))
									{
										sscanf(Parms[0],"%f",&NS);
									} else if ((tpt == 'S') || (tpt == 's')) {
										sscanf(Parms[0],"%f",&NS);
										NS *= -1;
									} else {
										NS = 0;
									}

									tpt = Parms[1][strlen(Parms[1])-1];
									Parms[1][strlen(Parms[1])-1] = 0;
									if ((tpt == 'E') || (tpt == 'e'))
									{
										sscanf(Parms[1],"%f",&EW);
									} else if ((tpt == 'W') || (tpt == 'w')) {
										sscanf(Parms[1],"%f",&EW);
										EW *= -1;
									} else {
										EW = 0;
									}

									Location_t templ;
									
									templ.x = (EW*10.0f+1020.0f)*24.0f;
									templ.y = (NS*10.0f-1020.0f)*24.0f;
									if (NumParms == 3)
										sscanf(Parms[2], "%f", &templ.z);
									else
										templ.z = 0.0f;
									
									templ.landblock = 
											( (BYTE) ((((DWORD) templ.x%192)/24)*8) + (((DWORD) templ.y%192)/24) ) |
											( (0x00) << 8) |
											( (BYTE) (templ.y/192.0f) << 16) |
											( (BYTE) (templ.x/192.0f) << 24)
										;
									templ.landblock -= 0x00010000;

									templ.x = (float) ((int) templ.x % 192);
									templ.y = (float) ((int) templ.y % 192);

									MoveUser(UserParsing, &templ, true, false);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !telemap syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"srvmsg") == 0) {
								try {
									sprintf(sayit, "Server Message from %s: %s", ConnUser[UserParsing].Char->Name, Contents);
									ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !srvmsg syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"who") == 0) {
								//Send list of people
								ServerMessage(UserParsing, "Listing Players:", COLOR_GREEN);
								for (int i=0;i<MaxUsers;i++)
								{
									if ((ConnUser[i].Connected) && (ConnUser[i].Char))
									{
										sprintf(sayit, "* %i - %i.%i.%i.%i:%i - %s - %s",
											i,
											ConnUser[i].sockaddy.sin_addr.S_un.S_un_b.s_b1,
											ConnUser[i].sockaddy.sin_addr.S_un.S_un_b.s_b2,
											ConnUser[i].sockaddy.sin_addr.S_un.S_un_b.s_b3,
											ConnUser[i].sockaddy.sin_addr.S_un.S_un_b.s_b4,
											ntohs(ConnUser[i].sockaddy.sin_port),
											ConnUser[i].Char->Name,
											ConnUser[i].Status);
										ServerMessage(UserParsing, sayit, COLOR_GREEN);
									}
								}
							} else if (stricmp(&command[1],"kick") == 0) {
								try {	
									DWORD tokick;
									sscanf(Contents, "%i", &tokick);
									if ((tokick < 128) && (ConnUser[tokick].Connected))
									{
										DisconnectUser(tokick);
									} else {
										//Try by name...
										for (int ii=0;ii<MaxUsers;ii++)
										{
											if (stricmp(Contents, ConnUser[ii].Char->Name) == 0)
											{	DisconnectUser(ii); break; }
										}
									}
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !kick syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"help") == 0) {
								ServerMessage(UserParsing, "List of KillaServ Commands:", COLOR_BLUE);
								ServerMessage(UserParsing, "-=* Help/Info Commands *=-", COLOR_YELLOW);
								ServerMessage(UserParsing, "* !help - This List\n \
															* !who - List of Users Online (Gets Player Numbers)", COLOR_GREEN);
								ServerMessage(UserParsing, "-=* Teleporting Commands *=-", COLOR_YELLOW);
								ServerMessage(UserParsing, "* !tele [X] [Y] [Z] [Landblock] - Teleport by Location_T\n \
															* !telemap [xx.xN/S] [xx.xE/W] [(Z)] - Teleport by NSEW coords with optional Z\n \
															* !telemapother [player number] [xx.xN/S] [xx.xE/W] [(Z)] - Teleports [player number] by NSEW coords with optional Z\n \
															* !goto [player number] - Teleports you to [player number]\n \
															* !summon [player number] - Summons [player number] to you", COLOR_GREEN);
								ServerMessage(UserParsing, "-=* Fun Commands *=-", COLOR_YELLOW);
								ServerMessage(UserParsing, "* !anim [anim number (0 to FF, hex)] - Make character do animation\n \
															* !particle [anim number (0 to FF, hex)] - Make particle effect emanete from character\n \
															* !smite [player number] - Hmmm...\n \
															* !changemodel [new model #] - Change your character to this model #", COLOR_GREEN);
								ServerMessage(UserParsing, "-=* Server Commands *=-", COLOR_YELLOW);
								ServerMessage(UserParsing, "* !givexp [player number] [Type] [Amount] - Give player XP\n \
															* !srvmsg [message] - Server Message\n \
															* !kick [player number] - Kick player by number\n \
															* !create [named item] - Create Named Item. Use '!create help' for a list of creatable objects", COLOR_GREEN);
								ServerMessage(UserParsing, "Use commands with care, no syntax checking is implemented yet.  It's easy to crash the server misusing commands...", COLOR_RED);
							} else if (stricmp(&command[1],"changemodel") == 0) {
								WORD NewModel;
								sscanf(Parms[0], "%i", &NewModel);
								ConnUser[UserParsing].Char->Model = NewModel;
								ServerMessage(UserParsing, "Model changed!  Relog to see the effect...", COLOR_RED);
							} else if (stricmp(&command[1],"givexp") == 0) {
								int PlayerFor;
								sscanf(Parms[0], "%i", &PlayerFor);
								int Type;
								sscanf(Parms[1], "%i", &Type);
								DWORD AmounttoGive;
								sscanf(Parms[2], "%u", &AmounttoGive);

								GiveXP(PlayerFor, Type, AmounttoGive);
							} else if (stricmp(&command[1],"damage") == 0) {
								BYTE DamagePacket[100], *PackPointer = DamagePacket;

								DWORD head = 0x1b2; memcpy(PackPointer, &head, 4); PackPointer += 4;
								int strlen; GenString("Someone", &strlen, PackPointer);	PackPointer += strlen;

								DWORD damage = 4; memcpy(PackPointer, &damage, 4); PackPointer += 4;
								double severity = 0.5f; memcpy(PackPointer, &severity, 8); PackPointer += 8;
								DWORD amount = rand() & 7; memcpy(PackPointer, &amount, 4); PackPointer += 4;

								Send200(DamagePacket, (DWORD) PackPointer - (DWORD) DamagePacket, 4, UserParsing);
							
							} else if (stricmp(&command[1],"create") == 0) {
								DWORD tpguid; bool freeg = false;
								while (!freeg)
								{
									tpguid = 0x78000000 | ((rand()*rand()) & 0x00FFFFFF);
									if (ObjectList[tpguid] == 0)
										freeg = true;
								}

								if (!Contents)
								{
									ServerMessage(UserParsing, "Try !create help...", COLOR_BLUE);
									return;
								}

								if (stricmp(Contents, "help") == 0)
								{
									ServerMessage(UserParsing, "!create currently supports the following objects: lifestone, obj, portal, portalloc.", COLOR_BLUE);
									ServerMessage(UserParsing, "+-> Lifestone - Creates a lifestone in front of you\n \
																+-> Obj - Creates generic objects... See !create obj help\n \
																+-> Portal [name] [loc] - Creates portal by //loc address, [loc] is same format as !tele\n \
																\\-> Portalloc [name] [loc] - Creates portal by NSEW coords, [loc] is same format as !telemap", COLOR_GREEN);
								} else if (stricmp(Parms[0], "portalloc") == 0) {
									cWorldObject_Portal *tpprt = new cWorldObject_Portal();
									
									tpprt->GUID = tpguid;
									memcpy(&tpprt->Loc, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));
									tpprt->Loc.x += 5.0f*ConnUser[UserParsing].PointX;
									tpprt->Loc.y += 5.0f*ConnUser[UserParsing].PointY;

									try {
										if (NumParms < 4)
										{
											ServerMessage(UserParsing, "Error in !create portalloc syntax, check !help!", COLOR_RED);
											return;
										}
										char *usrep = strchr(Parms[1],'_');
										while (usrep)
										{
											*usrep = ' ';
											usrep = strchr(Parms[1],'_');
										}
										strcpy(tpprt->Name, Parms[1]);
										
										float NS, EW;//, Nx, Ny;
										char tpt;
										tpt = Parms[2][strlen(Parms[2])-1];
										Parms[2][strlen(Parms[2])-1] = 0;
										if ((tpt == 'N') || (tpt == 'n'))
										{
											sscanf(Parms[2],"%f",&NS);
										} else if ((tpt == 'S') || (tpt == 's')) {
											sscanf(Parms[2],"%f",&NS);
											NS *= -1;
										} else {
											NS = 0;
										}

										tpt = Parms[3][strlen(Parms[3])-1];
										Parms[3][strlen(Parms[3])-1] = 0;
										if ((tpt == 'E') || (tpt == 'e'))
										{
											sscanf(Parms[3],"%f",&EW);
										} else if ((tpt == 'W') || (tpt == 'w')) {
											sscanf(Parms[3],"%f",&EW);
											EW *= -1;
										} else {
											EW = 0;
										}

										tpprt->DestLoc.x = (EW*10.0f+1020.0f)*24.0f;
										tpprt->DestLoc.y = (NS*10.0f-1020.0f)*24.0f;
										if (NumParms == 5)
											sscanf(Parms[4], "%f", &tpprt->DestLoc.z);
										else
											tpprt->DestLoc.z = 500.0f;
										
										tpprt->DestLoc.landblock = 
												( (BYTE) ((((DWORD) tpprt->DestLoc.x%192)/24)*8) + (((DWORD) tpprt->DestLoc.y%192)/24) ) |
												( (0x00) << 8) |
												( (BYTE) (tpprt->DestLoc.y/192.0f) << 16) |
												( (BYTE) (tpprt->DestLoc.x/192.0f) << 24)
											;
										tpprt->DestLoc.landblock -= 0x00010000;

										tpprt->DestLoc.x = (float) ((int) tpprt->DestLoc.x % 192);
										tpprt->DestLoc.y = (float) ((int) tpprt->DestLoc.y % 192);
									} catch (char *) {
										ServerMessage(UserParsing, "Error in !create portalloc syntax, check !help...", COLOR_RED);
										return;
									}

									LBObjects[tpprt->Loc.landblock >> 16][tpprt->GUID] = (cWorldObject *) tpprt;
									Portals[tpprt->Loc.landblock >> 16][tpprt->GUID] = tpprt;
									ObjectList[tpguid] = (cWorldObject *) tpprt;

									BYTE TpP[500];
									int packlen = tpprt->GenerateCreatePacket(TpP);
									for (int iii=0;iii<MaxUsers;iii++)
										if (ConnUser[iii].Connected)
											Send200(TpP, packlen, 3, iii);
								} else if (stricmp(Parms[0], "portal") == 0) {
									cWorldObject_Portal *tpprt = new cWorldObject_Portal();
									
									tpprt->GUID = tpguid;
									memcpy(&tpprt->Loc, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));
									tpprt->Loc.x += 3.0f;

									try {
										if (NumParms < 6)
										{
											ServerMessage(UserParsing, "Error in !create portal syntax, check !help!", COLOR_RED);
											return;
										}
										char *usrep = strchr(Parms[1],'_');
										while (usrep)
										{
											*usrep = ' ';
											usrep = strchr(Parms[1],'_');
										}
										strcpy(tpprt->Name, Parms[1]);
										sscanf(Parms[2],"%f",&tpprt->DestLoc.x);
										sscanf(Parms[3],"%f",&tpprt->DestLoc.y);
										sscanf(Parms[4],"%f",&tpprt->DestLoc.z);
										sscanf(Parms[5],"%08X",&tpprt->DestLoc.landblock);
									} catch (char *) {
										ServerMessage(UserParsing, "Error in !create portal syntax, check !help...", COLOR_RED);
										return;
									}

									LBObjects[tpprt->Loc.landblock >> 16][tpprt->GUID] = (cWorldObject *) tpprt;
									Portals[tpprt->Loc.landblock >> 16][tpprt->GUID] = tpprt;
									ObjectList[tpguid] = (cWorldObject *) tpprt;

									BYTE TpP[500];
									int packlen = tpprt->GenerateCreatePacket(TpP);
									for (int iii=0;iii<MaxUsers;iii++)
										if (ConnUser[iii].Connected)
											Send200(TpP, packlen, 3, iii);
								} else if (stricmp(Parms[0], "lifestone") == 0) {
									cWorldObject_LS *tpls = new cWorldObject_LS();
									
									tpls->GUID = tpguid;
									tpls->Type = OBJECT_LIFESTONE;
									memcpy(&tpls->Loc, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));
									tpls->Loc.x += 1.0f;

									LBObjects[tpls->Loc.landblock >> 16][tpls->GUID] = (cWorldObject *) tpls;
									ObjectList[tpguid] = (cWorldObject *) tpls;

									BYTE TpP[500];
									int packlen = tpls->GenerateCreatePacket(TpP);
									for (int iii=0;iii<MaxUsers;iii++)
										if (ConnUser[iii].Connected)
										Send200(TpP, packlen, 3, iii);
								} else if (stricmp(Parms[0], "obj") == 0) {
									if (stricmp(Parms[1], "help") == 0)
									{
										ServerMessage(UserParsing, "Syntax: !create obj -id ** [-name **] [-selectable]", COLOR_RED);
										ServerMessage(UserParsing, "+-> -id *** - Gives the object a Model # of **", COLOR_RED);
										ServerMessage(UserParsing, "+-> -name *** - Names the object this", COLOR_RED);
										ServerMessage(UserParsing, "\\-> -selectable - Makes the object selectable", COLOR_RED);
										return;
									}

									cWorldObject_ObjectID *tpObj = new cWorldObject_ObjectID();

									tpObj->Flags = 0;
									sprintf(tpObj->Name, "Unnamed Object");
									tpObj->Model = 1;

									try {
										int CurParm = 1;

										while (CurParm < NumParms)
										{
											if (stricmp(Parms[CurParm],"-id") == 0)
											{
												sscanf(Parms[CurParm+1], "%i", &tpObj->Model);
												CurParm++;
											}
											else if (stricmp(Parms[CurParm],"-selectable") == 0)
											{
												tpObj->Flags |= BYOBJ_SELECTABLE;
											}
											else if (stricmp(Parms[CurParm],"-name") == 0)
											{
												sscanf(Parms[CurParm+1], "%s", &tpObj->Name);
												CurParm++;
											}
											CurParm++;
										}
									} catch (...) {
										sprintf(sayit, "Error, check syntax: %s", messagetext);
										ServerMessage(UserParsing, sayit, COLOR_RED);
									}

									sprintf(sayit, "Model Created with ID: %i", tpObj->Model);
									ServerMessage(UserParsing, sayit, COLOR_RED);

									tpObj->GUID = tpguid;
									tpObj->Type = OBJECT_OBJECTID;
									memcpy(&tpObj->Loc, &ConnUser[UserParsing].Char->Loc, sizeof(Location_t));
									tpObj->Loc.x += 5.0f*ConnUser[UserParsing].PointX;
									tpObj->Loc.y += 5.0f*ConnUser[UserParsing].PointY;

									LBObjects[tpObj->Loc.landblock >> 16][tpObj->GUID] = (cWorldObject *) tpObj;
									ObjectList[tpguid] = (cWorldObject *) tpObj;

									BYTE TpP[500];
									int packlen = tpObj->GenerateCreatePacket(TpP);
									for (int iii=0;iii<MaxUsers;iii++)
										if (ConnUser[iii].Connected)
											Send200(TpP, packlen, 3, iii);
								}
							} else if (stricmp(&command[1],"renamesel") == 0) {
								DWORD GUIDtoChange = ConnUser[UserParsing].SelectedItem;
								cWorldObject *newobj = ObjectList[GUIDtoChange];
								if (newobj)
								{
									sprintf(newobj->Name, "%s", Parms[0]);
									sprintf(sayit, "Object %08X renamed to %s!", GUIDtoChange, newobj->Name);
									ServerMessage(UserParsing, sayit, COLOR_BLUE);
								}
							} else if (stricmp(&command[1],"renameobject") == 0) {
								DWORD GUIDtoChange; sscanf(Parms[0], "%08x", &GUIDtoChange);
								cWorldObject *newobj = ObjectList[GUIDtoChange];
								if (newobj)
								{
									sprintf(newobj->Name, "%s", Parms[1]);
									sprintf(sayit, "Object %08X renamed to %s!", GUIDtoChange, newobj->Name);
									ServerMessage(UserParsing, sayit, COLOR_BLUE);
								}
							} else if (stricmp(&command[1],"listobjects") == 0) {
								ServerMessage(UserParsing, "Listing Objects in Current Landblock...", COLOR_BLUE);
								for (std::map<DWORD, cWorldObject *>::iterator i = LBObjects[ConnUser[UserParsing].Char->Loc.landblock >> 16].begin(); i != LBObjects[ConnUser[UserParsing].Char->Loc.landblock >> 16].end(); i++)
								{
									if ((i->second->Type) && (i->second->GUID != ConnUser[UserParsing].Char->GUID))
									{
										sprintf(sayit, "Object: %08X, Type: %i, Name: %s", i->second->GUID, i->second->Type, i->second->Name);
										ServerMessage(UserParsing, sayit, COLOR_BLUE);
									}
								}
								ServerMessage(UserParsing, "End of list!", COLOR_BLUE);
							} else if (stricmp(&command[1],"anim") == 0) {
								try {
									DWORD tpdw;
									sscanf(Contents, "%04X",  &tpdw);
									
									SendAnim(UserParsing, tpdw);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !anim syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"particle") == 0) {
								try {
									DWORD tpdw;
									sscanf(Contents,"%02x",&tpdw);

									SendParticle(UserParsing, tpdw);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !particle syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"smite") == 0) {
								try {
									DWORD Playertokill;
									sscanf(Contents, "%04X",  &Playertokill);
									if (Playertokill > (DWORD) MaxUsers)
										return;

									SendAnim(Playertokill, 0x08);

									sprintf(sayit, "%s smites %s!", ConnUser[UserParsing].Char->Name, ConnUser[Playertokill].Char->Name);
									ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !smite syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"summon") == 0) {
								DWORD ToAssist;
								try {
									sscanf(Contents, "%i", &ToAssist);
									
									MoveUser(ToAssist, &ConnUser[UserParsing].Char->Loc, true, false);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !summon syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"goto") == 0) {
								DWORD ToGoto;
								try {
									sscanf(Contents, "%i", &ToGoto);
									
									MoveUser(UserParsing, &ConnUser[ToGoto].Char->Loc, true, false);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !goto syntax, check !help...", COLOR_RED);
									return;
								}
							} else if (stricmp(&command[1],"telemapother") == 0) {
								try {
									if (NumParms < 3)
									{
										ServerMessage(UserParsing, "Error in !telemap syntax, check !help!", COLOR_RED);
										return;
									}

									DWORD Chartotele;
									sscanf(Parms[0],"%i",&Chartotele);
									
									float NS, EW;//, Nx, Ny;
									char tpt;
									tpt = Parms[0][strlen(Parms[1])-1];
									Parms[1][strlen(Parms[1])-1] = 0;
									if ((tpt == 'N') || (tpt == 'n'))
									{
										sscanf(Parms[1],"%f",&NS);
									} else if ((tpt == 'S') || (tpt == 's')) {
										sscanf(Parms[1],"%f",&NS);
										NS *= -1;
									} else {
										NS = 0;
									}

									tpt = Parms[2][strlen(Parms[2])-1];
									Parms[2][strlen(Parms[2])-1] = 0;
									if ((tpt == 'E') || (tpt == 'e'))
									{
										sscanf(Parms[2],"%f",&EW);
									} else if ((tpt == 'W') || (tpt == 'w')) {
										sscanf(Parms[2],"%f",&EW);
										EW *= -1;
									} else {
										EW = 0;
									}

									Location_t templ;
									
									templ.x = (EW*10.0f+1020.0f)*24.0f;
									templ.y = (NS*10.0f-1020.0f)*24.0f;
									if (NumParms == 4)
										sscanf(Parms[3], "%f", &templ.z);
									else
										templ.z = 500.0f;

									templ.landblock = 
											( (BYTE) ((((DWORD) templ.x%192)/24)*8) + (((DWORD) templ.y%192)/24) ) |
											( (0x00) << 8) |
											( (BYTE) (templ.y/192.0f) << 16) |
											( (BYTE) (templ.x/192.0f) << 24)
										;
									templ.landblock -= 0x00010000;

									templ.x = (float) ((int) templ.x % 192);
									templ.y = (float) ((int) templ.y % 192);

									MoveUser(Chartotele, &templ, true, false);
								} catch (char *) {
									ServerMessage(UserParsing, "Error in !telemapother syntax, check !help...", COLOR_RED);
									return;
								}
							}
						} else {
							//Actual text message... omgomgmog
							BYTE tpmessage[0x1000];	ZeroMemory(tpmessage, sizeof(tpmessage));
							BYTE * PacketPointer = &tpmessage[0];
								
							int tplen;
							DWORD tpn = 0x00000037; memcpy(PacketPointer, &tpn, 4); PacketPointer += 4;
							GenString(messagetext, &tplen, PacketPointer);	PacketPointer += tplen;
							GenString(&ConnUser[UserParsing].Char->Name[0], &tplen, PacketPointer);	PacketPointer += tplen;
												memcpy(PacketPointer, &ConnUser[UserParsing].Char->GUID, 4); PacketPointer += 4;
							tpn = 0x2;			memcpy(PacketPointer, &tpn, 4); PacketPointer += 4;
							
							for (int iii=0;iii<MaxUsers;iii++)
							{
								if ((ConnUser[iii].Connected) && (ConnUser[iii].State == 6))
									Send200(&tpmessage[0], (int) (PacketPointer - &tpmessage[0]), 4, iii);
							}
						}

//									delete [] messagetext;
					}
				} else if (MType2 == 0x01DF) {
					//Emote!
					char *messagetext = new char[tphdr->m_wTotalSize-14];
					memcpy(messagetext,&tPackPtr[14],tphdr->m_wTotalSize-14);

					BYTE tpmessage[0x1000];	ZeroMemory(tpmessage, sizeof(tpmessage));
					BYTE * PacketPointer = &tpmessage[0];
						
					int tplen;
					DWORD tpn = 0x000001E2;
					memcpy(PacketPointer, &tpn, 4); PacketPointer += 4;
					memcpy(PacketPointer, &ConnUser[UserParsing].Char->GUID, 4); PacketPointer += 4;
					GenString(&ConnUser[UserParsing].Char->Name[0], &tplen, PacketPointer);	PacketPointer += tplen;
					GenString(messagetext, &tplen, PacketPointer);	PacketPointer += tplen;

					for (int iii=0;iii<MaxUsers;iii++)
					{
						if ((ConnUser[iii].Connected) && (ConnUser[iii].State == 6))
							Send200(&tpmessage[0], (int) (PacketPointer - &tpmessage[0]), 4, iii);
					}
				} else if (MType2 == 0x01E1) {
					//Emote Text...
					char MsgText[100];
					sprintf(MsgText, "%s %s", ConnUser[UserParsing].Char->Name, &tPackPtr[14]);

					ServerMessage(-1*(UserParsing+1), &MsgText[0], COLOR_GREY);
				} else if (MType2 == 0x53) {
					//Toggle Combat Mode
					//4th dword: 1 = peace, 2 = war
					
					DWORD CombatMode; memcpy(&CombatMode, &tPackPtr[12], 4);
					if (CombatMode == 1)
					{
						SendAnim(UserParsing, 0x3b);
					} else {
						if (ConnUser[UserParsing].Char->Equipped[EQUIPPED_WEAPON] == 0)
						{
							//Punching
							SendAnim(UserParsing, 0x3c);
						} else {
//										cWorldObject_Weapon *WeaponUsing = (cWorldObject_Weapon *) ConnUser[UserParsing].ObjectCache[ConnUser[UserParsing].Char->Equipped[EQUIPPED_WEAPON]];
//										if (!WeaponUsing)
//											WeaponUsing = (cWorldObject_Weapon *) ObjectList[ConnUser[UserParsing].Char->Equipped[EQUIPPED_WEAPON]];
						}
					}
				} else if (MType2 == 0x1F) {
					//Dunno
					
				} else if (MType2 == 0x021E) {
					//Dunno
					
				} else if (MType2 == 0xC8) {
					//Eval
					//4th dword = GUID to eval

					DWORD dwEvalGUID; memcpy(&dwEvalGUID, &tPackPtr[12], 4);

					EvalObject(UserParsing, dwEvalGUID);

					ActionComplete(UserParsing);
				} else if (MType2 == 0x19) {
					//Move Item
					//4th dword = GUID to move
					//5th dword = GUID of item to move it into... Player for equipping, bag for putting it in.
					//6th dword = Index in item to put it in

					DWORD GUIDMove;
					memcpy(&GUIDMove, &tPackPtr[12],4);
/*
					DWORD Pack22[] = {
						0x00000022,
						GUIDMove
					};
					Send200((BYTE *) &Pack22[0], sizeof(Pack22), 4, UserParsing);
*/
					//F7B0/22
					DWORD MoveItemArr[] = {
						0x0000F7B0,
						ConnUser[UserParsing].Char->GUID,
						ConnUser[UserParsing].EventCount,
						0x00000022,
						0,
						0,
						0
					};
					memcpy(&MoveItemArr[4],&tPackPtr[12],3*4);
					ConnUser[UserParsing].EventCount++;
					Send200((BYTE *) &MoveItemArr[0], sizeof(MoveItemArr), 4, UserParsing);

					if (GUIDMove == ConnUser[UserParsing].Char->Equipped[EQUIPPED_WEAPON])
					{
						//First 22D
						BYTE Pack22D[0x15]; ZeroMemory(Pack22D, 0x15);
						DWORD tpdword;	BYTE tpbyte;
						tpdword = 0x22D; memcpy(&Pack22D[0], &tpdword, 4);		//22D = set wielder/container
						tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack22D[4], &tpbyte, 1);	//EquipCount
						//byte here...
										 memcpy(&Pack22D[5], &GUIDMove, 4);		//Item
						tpdword = 3;     memcpy(&Pack22D[9], &tpdword, 4);		//Set Wielder
						tpdword = 0;     memcpy(&Pack22D[13], &tpdword, 4);		//Wielder = 0
						tpdword = ConnUser[UserParsing].EquipCount + 1;     memcpy(&Pack22D[17], &tpdword, 4);		//unknown2...
						Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);

						//229
						BYTE Pack229[0x11]; ZeroMemory(Pack229, 0x11);
						tpdword = 0x229; memcpy(&Pack229[0], &tpdword, 4);		//22D = set wielder/container
						tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack229[4], &tpbyte, 1);	//EquipCount
						//byte here...
										 memcpy(&Pack229[5], &GUIDMove, 4);		//Item
						tpdword = 0xA;   memcpy(&Pack229[9], &tpdword, 4);		//Unknown, always A...
						tpdword = 0;     memcpy(&Pack229[13], &tpdword, 4);		//Coverage = 0
						Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);

						//Second 22D
						ZeroMemory(Pack22D, 0x15);
						tpdword = 0x22D; memcpy(&Pack22D[0], &tpdword, 4);		//22D = set wielder/container
						tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack22D[4], &tpbyte, 1);	//EquipCount
						//byte here...
										 memcpy(&Pack22D[5], &GUIDMove, 4);		//Item
						tpdword = 2;     memcpy(&Pack22D[9], &tpdword, 4);		//Set Container
										 memcpy(&Pack22D[13], &tPackPtr[16], 4);		//Container = Bag
						tpdword = 0x34500;     memcpy(&Pack22D[17], &tpdword, 4);		//unknown2...
						Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);
						
						//EQ++
						ConnUser[UserParsing].EquipCount ++;
					} else {
						DWORD MoveItemArr[] = {
							0x0000F7B0,
							ConnUser[UserParsing].Char->GUID,
							ConnUser[UserParsing].EventCount,
							0x00000022,
							0,
							0,
							0
						};
						memcpy(&MoveItemArr[4],&tPackPtr[12],3*4);
						ConnUser[UserParsing].EventCount++;
						Send200((BYTE *) &MoveItemArr[0], sizeof(MoveItemArr), 4, UserParsing);

						BYTE Pack22D[0x15]; ZeroMemory(Pack22D, 0x15);
						DWORD tpdword;	BYTE tpbyte;
						tpdword = 0x22D; memcpy(&Pack22D[0], &tpdword, 4);		//22D = set wielder/container
						tpbyte = ConnUser[UserParsing].MoveItemCount;	memcpy(&Pack22D[4], &tpbyte, 1);	//MoveItemCount
						//byte here...
										 memcpy(&Pack22D[5], &GUIDMove, 4);		//Item
						tpdword = 2;     memcpy(&Pack22D[9], &tpdword, 4);		//2 = Set Container
										 memcpy(&Pack22D[13], &tPackPtr[16], 4);	//Container (pack)
						tpdword = 0x34500;     memcpy(&Pack22D[17], &tpdword, 4);		//unknown2
						Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);
						ConnUser[UserParsing].MoveItemCount += 2;
					}

//								DWORD abc[3];	memcpy(abc, &tPackPtr[12],4*3);
//								sprintf(sayit, "Move Item - Item: %08X, Into: %08X, Index: %08X",abc[0], abc[1], abc[2]); LogDisp(sayit);
				} else if (MType2 == 0x1A) {
					//Item Worn
					//4th dword = GUID of item
					//5th dword = coverage

					//We'll just go with equipped items now, m'kay? m'kay.

					DWORD GUIDEquip, Coverage;
					memcpy(&GUIDEquip, &tPackPtr[12],4);
					memcpy(&Coverage, &tPackPtr[16],4);		//Ignore coverage for now m'kay? m'kay.

					//F7B0/23
					DWORD F7B023[] = {
						0x0000F7B0,
						ConnUser[UserParsing].Char->GUID,
						ConnUser[UserParsing].EventCount,
						0x00000023,
						GUIDEquip,
						Coverage
					};
					ConnUser[UserParsing].EventCount++;
					Send200((BYTE *) &F7B023[0], sizeof(F7B023), 4, UserParsing);
					
					//First 22D
					BYTE Pack22D[0x15]; ZeroMemory(Pack22D, 0x15);
					DWORD tpdword;	BYTE tpbyte;
					tpdword = 0x22D; memcpy(&Pack22D[0], &tpdword, 4);		//22D = set wielder/container
					tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack22D[4], &tpbyte, 1);	//EquipCount
					//byte here...
									 memcpy(&Pack22D[5], &GUIDEquip, 4);		//Item
					tpdword = 3;     memcpy(&Pack22D[9], &tpdword, 4);		//Set Wielder
									 memcpy(&Pack22D[13], &ConnUser[UserParsing].Char->GUID, 4);		//Wielder = Player
					tpdword = ConnUser[UserParsing].EquipCount + 1;     memcpy(&Pack22D[17], &tpdword, 4);		//EquipCount + 1
					Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);
					
					//Second 22D
					ZeroMemory(Pack22D, 0x15);
					tpdword = 0x22D; memcpy(&Pack22D[0], &tpdword, 4);		//22D = set wielder/container
					tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack22D[4], &tpbyte, 1);	//EquipCount
					//byte here...
									 memcpy(&Pack22D[5], &GUIDEquip, 4);		//Item
					tpdword = 2;     memcpy(&Pack22D[9], &tpdword, 4);		//Set Container
					tpdword = 0;     memcpy(&Pack22D[13], &tpdword, 4);		//Container = 0
					tpdword = ConnUser[UserParsing].EquipCount + 1;     memcpy(&Pack22D[17], &tpdword, 4);		//EquipCount + 1
					Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);

					//229
					BYTE Pack229[0x11]; ZeroMemory(Pack229, 0x11);
					tpdword = 0x229; memcpy(&Pack229[0], &tpdword, 4);		//22D = set wielder/container
					tpbyte = ConnUser[UserParsing].EquipCount;	memcpy(&Pack229[4], &tpbyte, 1);	//EquipCount
					//byte here...
									 memcpy(&Pack229[5], &GUIDEquip, 4);		//Item
					tpdword = 0xA;   memcpy(&Pack229[9], &tpdword, 4);		//Unknown, always A...
									 memcpy(&Pack229[13], &Coverage, 4);		//Coverage = 0
					Send200(Pack22D, sizeof(Pack22D), 4, UserParsing);
					
					//F749
					DWORD F749[] = {
						ConnUser[UserParsing].Char->GUID,	//Wielder
						GUIDEquip,						//Item
						1,								//Unk1
						1,								//Unk2
						//40254,							//Unk3
						0x00254 | (0x40000 + 0x30000*((ConnUser[UserParsing].EquipCount - 1) / 2)),	//Unk3, try this...
						0x80000000						//Unk4
					};
					Send200((BYTE *) &F749[0], sizeof(F749), 4, UserParsing);

					//EQ++
					ConnUser[UserParsing].EquipCount ++;

					ConnUser[UserParsing].Char->Equipped[EQUIPPED_WEAPON] = GUIDEquip;

				} else if (MType2 == 0x1B) {
					//Item Dropped on Ground
					//4th dword = GUID dropped
					
					//?
				} else if (MType2 == 0x1BF) {
					//Item Selected
					//4th dword = GUID selected
					
					memcpy(&ConnUser[UserParsing].SelectedItem, &tPackPtr[12],4);
				} else if (MType2 == 0x01B7) {
					//Beginning of Jump
				} else if (MType2 == 0xF61B) {
					//Player Jumped
				} else if (MType2 == 0xCA) {
					//Dunno...
				} else if (MType2 == 0x01A1) {
					//Dunno...
				} else {
					char *tps2;
					sprintf(sayit, "* 200 Pack: Seq:%08X Srv:%08X CRC:%08X Size:%04X Tm:%02X Table:%02X", tphdr->m_dwSequence, tphdr->m_dwServerID, tphdr->m_dwCRC, tphdr->m_wTotalSize, tphdr->m_bTime, tphdr->m_bTable); LogDisp(sayit);
					
					char * tps3 = new char[1 + (amountRead - (sizeof(cClientPacketHeader)))*5];
					ZeroMemory(tps3,1 + (amountRead - (sizeof(cClientPacketHeader)))*5);
					tps2 = tps3;
					for (unsigned int tpl=0; tpl < (amountRead - (sizeof(cClientPacketHeader))); tpl++)
					{
						sprintf(tps2, "0x%02X ", tPackPtr[tpl]);
						tps2+=5;
					}
					sprintf(sayit, "\\>Data: %s", tps3); LogDisp(sayit);
					delete [] tps3;
				}
			} else {
					char *tps2;
					sprintf(sayit, "* 200 Pack: Seq:%08X Srv:%08X CRC:%08X Size:%04X Tm:%02X Table:%02X", tphdr->m_dwSequence, tphdr->m_dwServerID, tphdr->m_dwCRC, tphdr->m_wTotalSize, tphdr->m_bTime, tphdr->m_bTable); LogDisp(sayit);
					
					char * tps3 = new char[1 + (amountRead - (sizeof(cClientPacketHeader)))*5];
					ZeroMemory(tps3,1 + (amountRead - (sizeof(cClientPacketHeader)))*5);
					tps2 = tps3;
					for (unsigned int tpl=0; tpl < (amountRead - (sizeof(cClientPacketHeader))); tpl++)
					{
						sprintf(tps2, "0x%02X ", tPackPtr[tpl]);
						tps2+=5;
					}
					sprintf(sayit, "\\>Data: %s", tps3); LogDisp(sayit);
					delete [] tps3;
			}
			tPackPtr += tphdr->m_wTotalSize;
		}
	} else if (tphdr->m_dwFlags == 0x00001002) {
		//bTime Updated Client-Side, Ignore...
	} else {
		sprintf(sayit, "In:Seq:%08X Flg:%08X Srv:%08X CRC:%08X Size:%04X Tm:%02X Table:%02X", tphdr->m_dwSequence, tphdr->m_dwFlags, tphdr->m_dwServerID, tphdr->m_dwCRC, tphdr->m_wTotalSize, tphdr->m_bTime, tphdr->m_bTable); LogDisp(sayit);
		if ((int) (amountRead-sizeof(cClientPacketHeader)) > 0)
		{
			char * tps = new char[1 + (amountRead-sizeof(cClientPacketHeader))*5], *tps2 = tps;
			for (unsigned int tpl=0; tpl < (amountRead-sizeof(cClientPacketHeader)) ;tpl++)
			{
				sprintf(tps2, "0x%02X ", TestBuf[tpl+sizeof(cClientPacketHeader)]);
				tps2+=5;
			}
			sprintf(sayit, "* Data: %s",tps); LogDisp(sayit);
			delete [] tps;
		}
	}
}

void cACServer::RemoveItem(DWORD GUID)
{
	DWORD F747[] = {
		0x0000F747,
		GUID
	};
	for (int i=0;i<MaxUsers;i++)
	{
		if ((ConnUser[i].Connected) && (ConnUser[i].State == 6))
			Send200((BYTE *) F747, sizeof(F747), 3, i);
	}
}

void cACServer::EvalObject(int User, DWORD GUID)
{
	cWorldObject *cwoItem = ObjectList[GUID];

	BYTE OutPack[0x500];
	BYTE *PackPointer = OutPack;
	DWORD tpd;//	WORD tpw;
	
	//F7b0 header
	tpd = 0xF7B0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].Char->GUID, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].EventCount, 4);	PackPointer += 4;
	tpd = 0x00C9;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//C9 = Identify Object
	ConnUser[User].EventCount++;

	//Figure out Flags...	
	switch (cwoItem->Type)
	{
	case OBJECT_CHARACTER:
		tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Flags - Char/Monster
		break;
	case OBJECT_PORTAL:
		tpd = 0x00010000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Flags - Portal
		break;
	}

	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Successful Assess

	//Build Packet
	if (cwoItem->Type == OBJECT_CHARACTER)
	{
		
	}
	else if (cwoItem->Type == OBJECT_PORTAL)
	{
		cWorldObject_Portal *tpp = (cWorldObject_Portal *) cwoItem;
		DWORD Lower, Upper;
		Lower = (tpp->Flags & 0x00FF0000) >> 16;
		if (!Lower) Lower = 0xFFFFFFFF;
		Upper = (tpp->Flags & 0xFF000000) >> 24;
		if (!Upper) Upper = 0xFFFFFFFF;

		memcpy(PackPointer, &Lower, 4);	PackPointer += 4;		//Lower Limit
		memcpy(PackPointer, &Upper, 4);	PackPointer += 4;		//Upper Limit

		int packlen;
		GenString(tpp->Name, &packlen, PackPointer);  PackPointer += packlen;
	}
	
	Send200(OutPack, (DWORD) PackPointer - (DWORD) OutPack, 3, User);
}