#include "stdafx.h"

//#include "Network.h"
#include "FixedPackets.h"

cACServer::cACServer()
{
	//Constructor

	Listening = false;

	WSADATA wsaData;
	WORD wVersionRequested = 0x0202;
	int err;

	err = WSAStartup(wVersionRequested, &wsaData); //for winsock startup

	for (int i=0;i<128;i++)
	{
		ConnUser[i].Connected = false;
		ConnUser[i].State = 0;
		ConnUser[i].Char = 0;
	}

	MaxUsers = 0;
	NumConnected = 0;

	NumStored = 0;
	for (i=0;i<16384;i++)
		IPStore[i].Enabled = false;

	NeedToDelete = true;

	LoadOptions();

//	CalcNextFreeGUID();
}

cACServer::~cACServer()
{
	//Destructor
	SaveObjects();

	for (int i=0;i<128;i++)
		if (ConnUser[i].Connected)
			DisconnectUser(i);

	WSACleanup(); // Always do this at the end of a winsock program
}

void cACServer::LoadOptions()
{
	HKEY keyout;
	RegCreateKey(HKEY_CURRENT_USER, "Software\\ACServer\\Server", &keyout);

	long outvallen;
	outvallen = 2048;
	if (RegQueryValue(keyout, "MOTD", MOTDb, &outvallen) != ERROR_SUCCESS)
	{
		sprintf(MOTDb, "This server's operator needs to set its MOTD in the options...\n");
	}
	outvallen = 64;
	if (RegQueryValue(keyout, "ServerName", ServerName, &outvallen) != ERROR_SUCCESS)
	{
		sprintf(ServerName, "Unnamed Server");
	}
	outvallen = 4;
	if (RegQueryValueEx(keyout, "ListenPort", 0, NULL, (BYTE *) &ListenPort, (DWORD *) &outvallen) != ERROR_SUCCESS)
	{
		ListenPort = 9002;
	}
	
	RegCloseKey(keyout);
}

void cACServer::SaveOptions()
{
	HKEY keyout;
	RegCreateKey(HKEY_CURRENT_USER, "Software\\ACServer\\Server", &keyout);

	RegSetValue(keyout, "MOTD", REG_SZ, MOTDb, strlen(MOTDb));
	RegSetValue(keyout, "ServerName", REG_SZ, ServerName, strlen(ServerName));
	RegSetValueEx(keyout, "ListenPort", 0, REG_DWORD, (const BYTE *) &ListenPort, 4);
	
	RegCloseKey(keyout);
}

void cACServer::StartListen()
{
	LoadObjects();

	Listening = true;

	srand(time(NULL));

	struct sockaddr_in localAddress;	//IP and port of the local machine
	localAddress.sin_family = AF_INET;
	localAddress.sin_addr.s_addr = inet_addr(LocalIP);
	localAddress.sin_port = htons(ListenPort);

	charSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	BOOL newopt = TRUE;
	setsockopt(charSocket, SOL_SOCKET, SO_DONTLINGER, (const char *) &newopt, sizeof(newopt));
	int newbuf = 0x4000;
	setsockopt(charSocket, SOL_SOCKET, SO_SNDBUF, (const char *) &newbuf, sizeof(newbuf));
	setsockopt(charSocket, SOL_SOCKET, SO_RCVBUF, (const char *) &newbuf, sizeof(newbuf));
	
	bind(charSocket, (struct sockaddr *)&localAddress, sizeof(sockaddr_in) );

	listen(charSocket, SOMAXCONN);

	sprintf(sayit, "ACServer Server Up!  Listening on port %i...", ListenPort); LogDisp(sayit);

	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_RESETCONTENT, 0, 0);
	char tosay[100];
	sprintf(tosay, "User\tIP:Port\tStatus"); SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_ADDSTRING, 0, (WPARAM) &tosay[0]);
	for (int i=0;i<128;i++)
	{
		sprintf(tosay, "%i\tN/A\tNot Connected", i);
		SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_ADDSTRING, 0, (WPARAM) &tosay[0]);
	}

	//Set up event notification on main listen socket
	int err = WSAAsyncSelect(charSocket, MainWnd, WM_USER+1, FD_ACCEPT);
}

void cACServer::LoadAccounts()
{	
	FILE *in = fopen("Password.dat","rb");
	if (!in)
	{
		NumPasses = 0;
		return;
	}
	
	fseek(in, 0, SEEK_END);
	int filelen = ftell(in);
	fseek(in, 0, SEEK_SET);

	NumPasses = (int) (filelen/sizeof(stAccounts));

	fread(ZonePasses, filelen, 1, in);

	for (int i=0;i<NumPasses;i++)
	{
		ZonePasses[i].LoggedIn = (WORD) -1;
	}

	fclose(in);
}

void cACServer::CalcNextFreeGUID()
{
	NextFreeCharGUID = 0x50000000;
	for (int i=0;i<NumPasses;i++)
	{
		for (int j=0;j<5;j++)
		{
			if (ZonePasses[i].Chars[j] >= NextFreeCharGUID)
				NextFreeCharGUID = ZonePasses[i].Chars[j] + 1;
		}
	}
}

void cACServer::SaveAccounts()
{
	FILE *out = fopen("Password.dat","wb");
	if (!out)
	{
		MessageBox(NULL, "Error on Password File Save!  Very bad!!", "Critical Error!", MB_OK);
		return;
	}

	fwrite(ZonePasses, sizeof(stAccounts)*NumPasses, 1, out);

	fclose(out);
}


void cACServer::AcceptUser()
{
	sockaddr_in tpsa; int tpsal = sizeof(tpsa);
	SOCKET tpsck = accept(charSocket, (sockaddr *) &tpsa, &tpsal);

	if (tpsck != SOCKET_ERROR)
	{
		for (int i=0;i<128;i++)
		{
			if (!ConnUser[i].Connected)
			{
				memcpy(&ConnUser[i].sockaddy,&tpsa,sizeof(sockaddr));
				ConnUser[i].Socket = tpsck;

				ConnUser[i].Connected = true;

				ConnUser[i].State = -1;

				ConnUser[i].RecvState = -1;

				int err = WSAAsyncSelect(tpsck, MainWnd, WM_USER+2+i, FD_READ);

				break;
			}
		}
	}
}


void cACServer::NewRead(int UserNum)
{
	int Ret;
	for ( ;; )
	{
		if (ConnUser[UserNum].RecvState == -1)
		{
			Ret = recv(ConnUser[UserNum].Socket, (char *) &ConnUser[UserNum].CurrentHeader, sizeof(EmuHeader), NULL);
			if (Ret == SOCKET_ERROR)
			{
				int ErrRet = WSAGetLastError();
				if (ErrRet == WSAEWOULDBLOCK)
				{
					//No more data, come back later...
					return;
				} else {
					//Eep bad...  Disconnected...?
					DisconnectUser(UserNum);
					return;
				}
			}

			ConnUser[UserNum].RecvState = 0;
		} else {
			Ret = recv(ConnUser[UserNum].Socket, (char *) ConnUser[UserNum].RecvBuff, ConnUser[UserNum].CurrentHeader.payloadsize_, NULL);
			if (Ret == SOCKET_ERROR)
			{
				int ErrRet = WSAGetLastError();
				if (ErrRet == WSAEWOULDBLOCK)
				{
					//No more data, come back later...
					return;
				} else {
					//Eep bad...  Disconnected...?
					DisconnectUser(UserNum);
					return;
				}
			}

			ConnUser[UserNum].RecvState += Ret;

			if (ConnUser[UserNum].RecvState == ConnUser[UserNum].CurrentHeader.payloadsize_)
			{
				ConnUser[UserNum].RecvState = -1;

				switch (ConnUser[UserNum].CurrentHeader.type_)
				{
				case EmuHeader::Authenticate:
					AuthenticateUser(UserNum);
					break;
				case EmuHeader::Payload:
					ParseMessage(UserNum);
					break;
				}
			}
		}
	}
}


void cACServer::AuthenticateUser(int UserParsing)
{
	BYTE *RecvBuf = ConnUser[UserParsing].RecvBuff;

	//Get l/p here...
	char tpLogin[20], tpPass[20];
	memcpy(tpLogin, RecvBuf, 20);
	memcpy(tpPass, RecvBuf+20, 20);
	
	bool LoginFound = false;

	int LoginRet = 2;
	//0 = Valid Existing Login
	//1 = Valid New Login
	//2 = Incorrect Password

	//Check for existing Login by that name, or else create new one...
	for (int h=0;h<NumPasses;h++)
	{
		if (strcmp(ZonePasses[h].Login,tpLogin) == 0)
		{
			//ID Found
			LoginFound = true;
			if (strcmp(ZonePasses[h].Pass,tpPass) == 0)
			{
				//Valid Password

				sprintf(sayit, "User %s logged in.", ZonePasses[h].Login); LogDisp(sayit);
				LoginRet = 0;

				ConnUser[UserParsing].ZoneNum = h;
			}
			h = NumPasses;
		}
	}
	
	if (!LoginFound)
	{
		memcpy(ZonePasses[NumPasses].Login, tpLogin, 20);
		memcpy(ZonePasses[NumPasses].Pass, tpPass, 20);
		ZeroMemory(ZonePasses[NumPasses].Chars, sizeof(ZonePasses[NumPasses].Chars));
		ZeroMemory(ZonePasses[NumPasses].Secs2Del, sizeof(ZonePasses[NumPasses].Secs2Del));
		ZeroMemory(ZonePasses[NumPasses].Names, sizeof(ZonePasses[NumPasses].Names));
		sprintf(sayit, "New User %s logged in.", ZonePasses[NumPasses].Login); LogDisp(sayit);
		LoginRet = 1;
		ZonePasses[NumPasses].LoggedIn = -1;

		ConnUser[UserParsing].ZoneNum = NumPasses;

		NumPasses++;
		SaveAccounts();
	}

	memcpy(ConnUser[UserParsing].ZoneName, ZonePasses[ConnUser[UserParsing].ZoneNum].Login, 20);
	
	//Return LoginRet here...
	//
	//!!!
	send(ConnUser[UserParsing].Socket, (char *) &LoginRet, 4, NULL);

	if (LoginRet == 2)
	{
		//Invalid Login...

		ConnUser[UserParsing].Connected = false;

		closesocket(ConnUser[UserParsing].Socket);

		return;
	}					

	ConnUser[UserParsing].State = 1;
	
	ConnUser[UserParsing].curSeq = 1;
	ConnUser[UserParsing].EventCount = 0;

	ConnUser[UserParsing].msgID = 1;

	ConnUser[UserParsing].bTime = 0;
	ConnUser[UserParsing].RecTime = 0;
	ConnUser[UserParsing].LastTime = 0;

	ConnUser[UserParsing].move_count = 1;

	ConnUser[UserParsing].SelectedItem = 0;

	ConnUser[UserParsing].PortalCount = 0;
	ConnUser[UserParsing].OverrideCount = 0;
	ConnUser[UserParsing].AnimCount = 0;
	ConnUser[UserParsing].MoveItemCount = 1;
	ConnUser[UserParsing].EquipCount = 0;
	ConnUser[UserParsing].AttackCount = 0;
	ZeroMemory(ConnUser[UserParsing].Count237, sizeof(ConnUser[UserParsing].Count237));
	ZeroMemory(ConnUser[UserParsing].Count244,3);
	
	ConnUser[UserParsing].Connable = true;
	ConnUser[UserParsing].ConnTimer = 0;

	UpdateCharInfo(UserParsing, "Connected...");

	if (UserParsing >= MaxUsers)
		MaxUsers = UserParsing + 1;
}


void cACServer::ParseMessage(int UserParsing)
{
	BYTE *RecvBuf = ConnUser[UserParsing].RecvBuff;

	cClientPacketHeader *tphdr = (cClientPacketHeader *) RecvBuf;
	memcpy(tphdr, RecvBuf, sizeof(cClientPacketHeader));

	int PackSize = ConnUser[UserParsing].CurrentHeader.payloadsize_;

	ConnUser[UserParsing].LastTime = ConnUser[UserParsing].RecTime;
	ConnUser[UserParsing].lastInSeq = tphdr->m_dwSequence;

	switch (tphdr->m_dwFlags)
	{
	case 0x00000002:	//Error
		sprintf(sayit, "* Error Packet - 2...");
		LogDisp(sayit); 
		break;
	case 0x00000004:	//Ping
		if (ConnUser[UserParsing].bTime < tphdr->m_bTime)
			ConnUser[UserParsing].bTime = tphdr->m_bTime;

		//Reply at some point...

		break;
	case 0x00000020:	//Disconnect
		DisconnectUser(UserParsing);
		break;
	case 0x00100000:	//Timing
		//Ignore
		break;
	case 0x00200000:	//Unknown
		//Ignore
		break;
	case 0x00000080:	//Connection Attempt
		if (ConnUser[UserParsing].Connable)
		{
			UpdateCharInfo(UserParsing, "Sending Initial Info...");
			
			ConnUser[UserParsing].Connable = false;
			ConnUser[UserParsing].ConnTimer = 5;
			
			SendCharList(UserParsing);
			
			UpdateCharInfo(UserParsing, "Selecting Character");
		}
		break;
	case 0x00000200:
		Parse200(UserParsing, *((DWORD *) (RecvBuf+sizeof(cClientPacketHeader))), &RecvBuf[sizeof(cClientPacketHeader) + 4], PackSize - sizeof(cClientPacketHeader) - 4);
		break;
	default:
		//Unknown Type
		char * tps = new char[1 + (PackSize-sizeof(cClientPacketHeader))*5], *tps2 = tps;
		for (unsigned int tpl=0; tpl < (PackSize-sizeof(cClientPacketHeader)); tpl++)
		{
			sprintf(tps2, "0x%02X ", RecvBuf+sizeof(cClientPacketHeader)+tpl);
			tps2+=5;
		}
		sprintf(sayit, "In:Seq:%08X Flg:%08X Srv:%08X CRC:%08X Size:%04X Tm:%02X Table:%02X * Data: %s", tphdr->m_dwSequence, tphdr->m_dwFlags, tphdr->m_dwServerID, tphdr->m_dwCRC, tphdr->m_wTotalSize, tphdr->m_bTime, tphdr->m_bTable, tps); LogDisp(sayit);
		delete [] tps;
		break;
	}
}

void cACServer::LogDisp(char *toDisp)
{
	int numtodisp = 1 + (int) (strlen(toDisp)/90);
	char tosay[91];
	tosay[90] = 0;
	int lognum;
	for (int i=0;i<numtodisp;i++)
	{
		memcpy(&tosay[0],toDisp+(i*90),90);
		lognum = SendDlgItemMessage(MainWnd, IDC_LIST1, LB_ADDSTRING/*LB_INSERTSTRING*/, 0, (LPARAM) tosay);
	}

	SendDlgItemMessage(MainWnd, IDC_LIST1, LB_SETTOPINDEX, lognum-16, 0);
}

void cACServer::IncTimes()
{
	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
		{
			if (!ConnUser[i].Connable)
			{
				ConnUser[i].ConnTimer--;
				if (!ConnUser[i].ConnTimer)
					ConnUser[i].Connable = true;
			}

//			ConnUser[i].bTime++;
//			if (ConnUser[i].bTime > 80)
//				ConnUser[i].bTime = 0;

			ConnUser[i].RecTime++;
			if ((ConnUser[i].RecTime - ConnUser[i].LastTime) > 30)		//If 30 seconds...
			{
				DisconnectUser(i);
//				sprintf(sayit, "* User %i timed out, disconnecting...", i);  LogDisp(sayit);
			}
		}
	}

	if (NeedToDelete)
	{
		NeedToDelete = false;
		for (i=0;i<NumPasses;i++)
		{
			for (int j=0;j<5;j++)
			{
				if (ZonePasses[i].Secs2Del[j])
				{
					ZonePasses[i].Secs2Del[j]--;
					if (!ZonePasses[i].Secs2Del[j])
					{
						ZonePasses[i].Chars[j] = 0;
						ZeroMemory(ZonePasses[i].Names[j], 32);
						if (ZonePasses[i].LoggedIn != -1)
							SendCharList(ZonePasses[i].LoggedIn);
						SaveAccounts();
					} else {
						NeedToDelete = true;
					}
				}

			}
		}
	}
}

void cACServer::SendCharList(int UserParsing)
{
	BYTE PlayerList[0x2000];
	BYTE *PackPointer = PlayerList;

	DWORD tpdword;

	tpdword = 0x0000F658;			memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
	tpdword = 0x00000000;			memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
	
	int NumChars = 0;
	for (int i=0;i<5;i++)
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i])
			NumChars++;
									memcpy(PackPointer, &NumChars, 4);	PackPointer += 4;

	int outlen;
	for (i=0;i<5;i++)
	{
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i])
		{
			tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x7FFFFFFF;
				memcpy(PackPointer,&tpdword,4); PackPointer += 4;	//GUID

			GenString(ZonePasses[ConnUser[UserParsing].ZoneNum].Names[i], &outlen, PackPointer); PackPointer += outlen;

			//Time remaining till perma-delete
			if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x80000000)
				tpdword = ZonePasses[ConnUser[UserParsing].ZoneNum].Secs2Del[i];
			else
				tpdword = 0;

					memcpy(PackPointer, &tpdword, 4);	PackPointer += 4;
		}
	}

	tpdword = 0;
	for (i=0;i<5;i++)
		if (ZonePasses[ConnUser[UserParsing].ZoneNum].Chars[i] & 0x80000000)
			tpdword++;
									memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//0 - DWORD
	tpdword = 5;					memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//5 - DWORD
	tpdword = 0xBEEFBEEF;			memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//BEEFBEEF - DWORD
	memcpy(PackPointer,&ConnUser[UserParsing].ZoneName[0],0x14);	PackPointer += 20;		//Zone Name - 20 Chars
	tpdword = 0;					memcpy(PackPointer,&tpdword,4);	PackPointer += 4;		//0 - DWORD
	Send200(PlayerList, (int) (PackPointer - PlayerList), 4, UserParsing);

//MOTD
	PackPointer = PlayerList;

	NumConnected = 0;
	for (int hh=0;hh<128;hh++)
	{
		if (ConnUser[hh].Connected)
			NumConnected++;
	}

	sprintf(MOTDa, "Hello, %s, welcome to %s!\nThis server is running a Beta copy of ACServer.  Check www.acserver.org for updates.\n\nThere are currently: %i Clients Connected.\n\n", ConnUser[UserParsing].ZoneName, ServerName, NumConnected);

	DWORD Zero = 0x0000F65A;				memcpy(PackPointer,&Zero,4);	PackPointer += 4;		//F65A - DWORD
	GenString(MOTDa, &outlen, PackPointer);	PackPointer += outlen;
	GenString(MOTDb, &outlen, PackPointer); PackPointer += outlen;

	Send200(PlayerList, (int) (PackPointer - PlayerList), 4, UserParsing);
}

void cACServer::Send200(BYTE *data, int datalen, WORD wUnk, int User)
{
	EmuHeader tph;
	tph.payloadsize_ = datalen;
	tph.message_ = wUnk;
	send(ConnUser[User].Socket, (char *) &tph, sizeof(tph), NULL);

	send(ConnUser[User].Socket, (char *) data, datalen, NULL);

//	for (int i=0;i<datalen;i+=464)	//Split into 464 byte chunks 4 some reason... :P
//		send(ConnUser[User].Socket, (char *) data+i, (datalen - i >= 464) ? 464 : (datalen - i), NULL);
}

void cACServer::DisconnectUser(int User)
{
	if (ConnUser[User].Connected == false)
		return;

	ConnUser[User].ObjectCache.clear();

	ConnUser[User].Connected = false;

	closesocket(ConnUser[User].Socket);

	int NumConn = 0;
	for (int hh=0;hh<128;hh++)
	{
		if (ConnUser[hh].Connected)
			NumConn++;
	}

	ZonePasses[ConnUser[User].ZoneNum].LoggedIn = (WORD) -1;

	UpdateCharInfo(User, "Not Connected");

	//Recalc MaxUsers
	for (int i=127;i>=0;i--)
	{
		if (ConnUser[i].Connected)
		{	MaxUsers = i+1;	break; }
	}

	if (ConnUser[User].Char == 0)
		return;

	SaveCharacter(User);

	LBObjects[ConnUser[User].Char->Loc.landblock >> 16].erase(ConnUser[User].Char->GUID);

	ObjectList.erase(ConnUser[User].Char->GUID);

	//Return him to char select screen
	DWORD tph3 = 0x0000F653;

	Send200((BYTE *) &tph3, sizeof(tph3), 4, User);

	//Disconnect Him
	ConnUser[User].Connected = false;

	sprintf(sayit, "Player %s Disconnected!", ConnUser[User].Char->Name);
	ServerMessage(GLOBAL_MESSAGE, sayit, COLOR_GREEN);

	//Delete him from other people
	RemoveItem(ConnUser[User].Char->GUID);

	ConnUser[User].Char = 0;
}

void cACServer::UpdateCharInfo(int User, char * NewText)
{
	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_DELETESTRING, User+1, 0);

	char TpStr[100];
	strcpy(ConnUser[User].Status, NewText);
	if (ConnUser[User].Connected)
	{
		sprintf(TpStr, "%i\t%i.%i.%i.%i:%i\t%s", User,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b1,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b2,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b3,
			ConnUser[User].sockaddy.sin_addr.S_un.S_un_b.s_b4,
			ntohs(ConnUser[User].sockaddy.sin_port),
			NewText);
	} else {
		sprintf(TpStr, "%i\tN/A\t%s", User, NewText);
	}
	SendDlgItemMessage(MainWnd, IDC_CHARLIST, LB_INSERTSTRING, User+1, (WPARAM) &TpStr[0]);
}

void cACServer::ServerMessage(int User, char *Text, DWORD Color)
{
	try {
		BYTE TpMes[1024]; BYTE * PackPtr = &TpMes[0];
		DWORD tpf;
		int outlen;
		tpf = 0x0000F62C;		memcpy(PackPtr,&tpf,4);	PackPtr += 4;
		GenString(Text, &outlen, PackPtr); PackPtr += outlen;
		tpf = Color;		memcpy(PackPtr,&tpf,4);	PackPtr += 4;
		if ((User >= 0) && (User < 128))
		{
			Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, User);
		} else if (User == GLOBAL_MESSAGE) {
			//Global Message...
			for (int i=0;i<MaxUsers;i++)
			{
				if ((ConnUser[i].Connected) && (ConnUser[i].State >= 5))
					Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, i);
			}
		} else if ((User > -128) && (User < 0)) {
			//Global Message except User
			for (int i=0;i<MaxUsers;i++)
			{
				if ((ConnUser[i].Connected) && (ConnUser[i].State >= 5) && (i != (User*-1)-1))
					Send200(&TpMes[0], (int) (PackPtr - &TpMes[0]), 4, i);
			}
		}
	} catch (char * ) {
		//Meh...
		
	}
}

void cACServer::SaveCharacter(int CharNum)
{
	FILE *charread = fopen("CharList.dat","r+b");

	fseek(charread, 0, SEEK_END);
	int numchars = ftell(charread)/(SIZEOFCHAR+68);
	fseek(charread, 0, SEEK_SET);

	DWORD GUIDtemp;
	for (int j=0;j<numchars;j++)
	{
		fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);
		fread(&GUIDtemp, 4, 1, charread);
		if (GUIDtemp == ConnUser[CharNum].Char->GUID)
		{
			fseek(charread, j*(SIZEOFCHAR+68), SEEK_SET);

			BYTE TPlog[5000]; int charsize = ConnUser[CharNum].Char->Serialize(TPlog);
			fwrite(&ConnUser[CharNum].Char->GUID, 4, 1, charread);
			fwrite(ConnUser[CharNum].Char->Name, 32, 1, charread);
			fwrite(&ConnUser[CharNum].Char->Loc, sizeof(Location_t), 1, charread);
			fwrite(TPlog, charsize, 1, charread);
			
//			fwrite(ConnUser[CharNum].Char, sizeof(cWorldObject_Char), 1, charread);
			j = numchars;
		}
	}

	fclose(charread);
}

int cACServer::Generate_F7B0_13(BYTE * OutPack, int User)
{
	BYTE *PackPointer = OutPack;
	DWORD tpd;	WORD tpw;
	
	//F7b0 header
	tpd = 0xF7B0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].Char->GUID, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].EventCount, 4);	PackPointer += 4;
	tpd = 0x0013;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//13 = Login Char
	ConnUser[User].EventCount++;
	
	//On the the actual packet... - loginmask1 - 7B in the stuff...
	tpd = 0x001B;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginMask1
	tpd = 0x000A;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown1 - A
	
	//count1
	tpw = 0x000C;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count for next section
	tpw = 0x0040;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown for this - 40

	tpd = 0x0081;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x02F7;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x0084;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	//Burden, calc later..
	tpd = 0x0005;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//5 = burden
	tpd =      1;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//1 = ?

	tpd = 0x008B;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0037;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	//Pyreals, calc later
	tpd = 0x0014;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//14 = pyreals
	tpd =      0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//0
	//Total Xp
	tpd = 0x0015;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//15 = total xp
						memcpy(PackPointer, &ConnUser[User].Char->TotalXP, 4);	PackPointer += 4;
	//Unassigned XP
	tpd = 0x0016;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//16 = unassigned xp
						memcpy(PackPointer, &ConnUser[User].Char->UnassignedXP, 4);	PackPointer += 4;
	//Unassigned Skill Points
	tpd = 0x0018;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//18 = unassigned skill points
						memcpy(PackPointer, &ConnUser[User].Char->SkillCredits, 4);	PackPointer += 4;
	//Level, calc later....
	tpd = 0x0019;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//19 = Level
						memcpy(PackPointer, &ConnUser[User].Char->Level, 4);	PackPointer += 4;
	//Rank, calc later
	tpd = 0x001e;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//1e = Rank
	tpd = 0x0000;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x0028;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;

	tpd = 0x002F;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	tpd = 0x0000;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
//end of that vector

	//count2	
	tpw = 0x0005;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count2
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown - 20
	//count2 vector here... i.e. nothing yet

	//Count2 Stuff...
	DWORD CHAR_DATA_VECTOR2[] = {
		0x00000004, 0x00000000, 0x0000002C, 0x00000000, 
		0x0000002D, 0x00000000, 0x0000002E, 0x00000000, 
		0x0000002F, 0x00000000, 
	};
	memcpy(PackPointer, CHAR_DATA_VECTOR2, sizeof(CHAR_DATA_VECTOR2));	PackPointer += sizeof(CHAR_DATA_VECTOR2);

	//strings
	tpw = 0x0004;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//stringcount
	tpw = 0x0010;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//strings unknown - 10

	//Char Name
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						GenString(ConnUser[User].Char->Name, (int *) &tpd, PackPointer);	PackPointer += tpd;
	//Gender
	tpd = 0x0003;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	char Gender[7];
	if (ConnUser[User].Char->Gender == 0)
		sprintf(Gender, "Female");
//	if (ConnUser[User].Char->Gender == 1)
	else
		sprintf(Gender, "Male");
	GenString(Gender, (int *) &tpd, PackPointer);	PackPointer += tpd;

	//Race
	char Race[12];
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
	if (ConnUser[User].Char->Race == 0)
		sprintf(Race, "Aluvian");
	if (ConnUser[User].Char->Race == 1)
		sprintf(Race, "Gharun'Dim");
//	if (ConnUser[User].Char->Race == 2)
	else
		sprintf(Race, "Sho");
	GenString(Race, (int *) &tpd, PackPointer);	PackPointer += tpd;

	//Class	- "Server Tester" for now... ;-)
	tpd = 0x0005;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						GenString("Server Tester", (int *) &tpd, PackPointer);	PackPointer += tpd;

	//count4
	tpw = 0x0001;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count4
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count4unknown - 20
	//count4 vector here... i.e. nothing yet

	//sec4 stuff
	tpd = 0x0004;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//
	tpd = 0x30000000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//

	//Loginmask1 info goes here

	

	//End loginmask1 shit
	//should be 103, 100 = spellbook, 2 = skills
	tpd = 0x0103;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Loginmask2
	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown14 - always 1
	tpd = 0x01ff;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//attributemask - always 0x1ff
	
	//Primary Stats						
	for (int i=0;i<6;i++)
	{
		if (i == 3)
		{
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[2], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[2], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[2], 4);	PackPointer += 4;
		} else if (i == 2) {
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[3], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[3], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[3], 4);	PackPointer += 4;
		} else {
			memcpy(PackPointer, &ConnUser[User].Char->CurrentStat[i], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->InitialStat[i], 4);	PackPointer += 4;
			memcpy(PackPointer, &ConnUser[User].Char->XPIntoStat[i], 4);	PackPointer += 4;
		}
	}

	//Secondary Stats
	for (i=0;i<3;i++)
	{
		memcpy(PackPointer, &ConnUser[User].Char->SecondaryStatInc[i], 4);		PackPointer += 4;
		tpd = 0;	memcpy(PackPointer, &tpd, 4);		PackPointer += 4;	//?... always 0 afawct
		memcpy(PackPointer, &ConnUser[User].Char->XPIntoSecondaryStat[i], 4);		PackPointer += 4;
		memcpy(PackPointer, &ConnUser[User].Char->CurrentSecondaryStat[i], 4);		PackPointer += 4;
	}

	//Skillcount
	tpw = 0x0023;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//count
	tpw = 0x0020;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//unknown - 20
	for (i=1;i<40;i++)
	{
		if (ConnUser[User].Char->SkillTrain[i] > 0)
		{
			memcpy(PackPointer, &i, 4);	PackPointer += 4;		//Skill number
			memcpy(PackPointer, &ConnUser[User].Char->SkillInc[i], 4);	PackPointer += 4;		//Skill number
			memcpy(PackPointer, &ConnUser[User].Char->SkillTrain[i], 4);	PackPointer += 4;		//Trained Status
			memcpy(PackPointer, &ConnUser[User].Char->SkillXP[i], 4);	PackPointer += 4;		//Total XP into skill
			memcpy(PackPointer, &ConnUser[User].Char->SkillFreeXP[i], 4);	PackPointer += 4;	//Free XP at gen
			tpd = 0x0000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;	//unknown2
			
			//double thingy
			tpd = 0xE5C9D958;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
			tpd = 0x4171BA45;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
		}
	}

	//Loginmask2 time
	//100 - spellbook
	WORD numspells = 0;
	for (i=0;i<96;i++)
	{
		for (DWORD tpdd=ConnUser[User].Char->SpellsLearned[i];tpdd > 0;tpdd >>= 1)
		{
			if (tpdd & 1)
				numspells++;
		}
	}
						memcpy(PackPointer, &numspells, 2);	PackPointer += 2;		//count
	tpw = 0x0040;		memcpy(PackPointer, &tpw, 2);	PackPointer += 2;		//spellcountunknown

	int tpinc;	float tpcharge = 1.0f;
	for (i=0;i<96;i++)
	{
		tpinc = 0;
		for (DWORD TpSpell = ConnUser[User].Char->SpellsLearned[i]; TpSpell > 0; TpSpell >>= 1)
		{
			if (TpSpell & 1)
			{
				tpd = (i << 5) | tpinc;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//spell ID
											memcpy(PackPointer, &tpcharge, 4);	PackPointer += 4;		//charge
				tpd = 0;					memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown2
				tpd = 0;					memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown3
			}
			tpinc++;
		}
	}

	//200 - enchantments... whoooa there cowboy

	DWORD Numshorts = 0;
	for (i=0;i<10;i++)
		if (ConnUser[User].Char->ShortCut[i])
			Numshorts++;

	if (Numshorts)
		tpd = 0x0005;
	else
		tpd = 0x0004;
						memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3

	//4, character options
//	tpd = 0x01ECE563;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3
	tpd = 0x01E4E56A;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//loginmask3

	if (Numshorts)
	{
		//1, shortcut bar
		memcpy(PackPointer, &Numshorts, 4);	PackPointer += 4;
		for (i=0;i<10;i++)
			if (ConnUser[User].Char->ShortCut[i])
			{
				memcpy(PackPointer, &i, 4);	PackPointer += 4;									//Position
				memcpy(PackPointer, &ConnUser[User].Char->ShortCut[i], 4);	PackPointer += 4;	//Object
				tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;					//Unknown - 0
			}
	}

	//Spellbars
	for (i=0;i<5;i++)
	{
		DWORD numspells = 0;
		for (numspells=0;ConnUser[User].Char->SpellBar[i][numspells] > 0;numspells++) { }

		memcpy(PackPointer, &numspells, 4);	PackPointer += 4;
		for (int j=0;j<numspells;j++)
		{
			memcpy(PackPointer, &ConnUser[User].Char->SpellBar[i][j], 4);	PackPointer += 4;
		}
	}

	//comp buyer...
	
	//Inventory
	DWORD numinv = 0;
	for (i=0;i<7;i++)
		if (ConnUser[User].Char->SidePacks[i])
			numinv++;
	for (i=0;i<102;i++)
		if (ConnUser[User].Char->MainPack[i])
			numinv++;

	memcpy(PackPointer, &numinv, 4);	PackPointer += 4;
	for (i=0;i<7;i++)
		if (ConnUser[User].Char->SidePacks[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->SidePacks[i], 4);	PackPointer += 4;
			tpd = 1;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//IsContainer = 1
		}
	for (i=0;i<102;i++)
		if (ConnUser[User].Char->MainPack[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->MainPack[i], 4);	PackPointer += 4;
			tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//IsContainer = 0
		}

	//Equipped Shizzit
	DWORD numequip = 0;
	for (i=0;i<25;i++)
		if (ConnUser[User].Char->Equipped[i])
			numequip++;

	memcpy(PackPointer, &numequip, 4);	PackPointer += 4;
	for (i=0;i<25;i++)
		if (ConnUser[User].Char->Equipped[i])
		{
			memcpy(PackPointer, &ConnUser[User].Char->Equipped[i], 4);	PackPointer += 4;
			tpd = (1 << i);	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Coverage
			tpd = 0;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//unknown3
		}

	return((int) (PackPointer - OutPack));
}

void cACServer::LoadObjects()
{
	sprintf(sayit, "Loading Objects.Dat..."); LogDisp(sayit);

	FILE *in = fopen("Objects.dat","rb");
	if (!in)
	{
		sprintf(sayit, "Objects.Dat Empty!  Initializing with an empty empty world."); LogDisp(sayit);
		return;
	}

	BYTE ObjData[4096];
	cWorldObject *tpObj = (cWorldObject *) ObjData;
	
	int errtest;

	cWorldObject *tpo;

	while (!feof(in))
	{
		DWORD ObjType;  errtest = fread(&ObjType, 4, 1, in);
	
		if (!errtest)
		{
			fclose(in);
			sprintf(sayit, "Objects.Dat Loaded!"); LogDisp(sayit);
			return;
		}

		switch (ObjType)
		{
		case OBJECT_LIFESTONE:
			tpo = new cWorldObject_LS();
			break;
		case OBJECT_OBJECTID:
			tpo = new cWorldObject_ObjectID();
			break;
		case OBJECT_PORTAL:
			tpo = new cWorldObject_Portal();
			break;
		case OBJECT_CHARACTER:
			//Ignore
			MessageBox(NULL, "Fatal Error: Character found in objects.dat...", "Fatal Error", MB_OK);
			break;
		default:
			//eep... bad.
			MessageBox(NULL, "Fatal Error: Unknown Object type in Objects.Dat.  Please Restore Objects.Dat.  Objects.Dat loading halted...", "Fatal Error", MB_OK);
			return;
		}

		if (ObjType != OBJECT_CHARACTER)
		{
			fread(&tpo->GUID, 4, 1, in);
			fread(tpo->Name, 32, 1, in);
			fread(&tpo->Loc, 32, 1, in);

			fread(ObjData, tpo->Size, 1, in);
			tpo->LoadFrom(ObjData);
			ObjectList[tpo->GUID] = tpo;
			LBObjects[tpo->Loc.landblock >> 16][tpo->GUID] = tpo;
		}

		if (ObjType == OBJECT_PORTAL)
			Portals[tpo->Loc.landblock >> 16][tpo->GUID] = (cWorldObject_Portal *) tpo;
	}
}

void cACServer::SaveObjects()
{
//	DeleteFile("Objects.Bak");
//	rename("Objects.Dat", "Objects.Bak");

	FILE *out = fopen("Objects.Dat","wb");
	if (!out)
	{
		//bad...
		MessageBox(NULL, "Can't open objects.dat for writing... very bad...", "Fatal Error", MB_OK);
		return;
	}
	
	BYTE SerData[4096];	int SerSize;

	for (std::map<DWORD, cWorldObject *>::iterator i = ObjectList.begin(); i != ObjectList.end(); i++)
	{
		if ((i->second) && (i->second->Type != OBJECT_CHARACTER))
		{
			fwrite(&i->second->Type, 4, 1, out);
			fwrite(&i->second->GUID, 4, 1, out);
			fwrite(i->second->Name, 32, 1, out);
			fwrite(&i->second->Loc, 32, 1, out);

			SerSize = i->second->Serialize(SerData);
			if (SerSize)
				fwrite(SerData, SerSize, 1, out);
		}
	}

	fclose(out);
}

void cACServer::SendLandblock(int User, WORD Landblock)
{
	int packlen;
	BYTE TpP[500];
	
	for (std::map<DWORD, cWorldObject *>::iterator i = LBObjects[Landblock].begin(); i != LBObjects[Landblock].end(); i++)
	{
		if ((i->second->Type != 0) && (i->second->GUID != ConnUser[User].Char->GUID))
		{
			packlen = i->second->GenerateCreatePacket(TpP);
			Send200(TpP, packlen, 3, User);
		}
	}
}

void cACServer::MoveUser(int User, Location_t *loc, bool Portal, bool Override)
{
	BYTE MOVE_ENTITY[] = {
		0x48, 0xF7, 0x00, 0x00, 0x03, 0x40, 0x16, 0x78, 0x34, 0x00, 0x00, 0x00, 0x08, 0x01, 0x64, 0x81,   //(H___@_x4_____d_) - 0000
		0xA1, 0xDB, 0x18, 0x41, 0xF0, 0x07, 0x19, 0x43, 0xD7, 0xA3, 0xBD, 0x41, 0xAA, 0x77, 0x6D, 0x3F,   //(___A___C___A_wm?) - 0010
		0xE0, 0x42, 0xBF, 0xBE, 0x39, 0x06, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,                           //(_B__9_______)     - 0020
	};

	for (std::map<DWORD, cWorldObject *>::iterator i = Portals[loc->landblock >> 16].begin(); i != Portals[loc->landblock >> 16].end(); i++)
	{
		if (GetDistance(loc, &i->second->Loc) < 0.01)
		{
			MoveUser(User, &((cWorldObject_Portal *)i->second)->DestLoc, true, false);
//			ServerMessage(User, "Entered Portal... Teleporting...", COLOR_RED);
		}
	}

	float heading = acos(loc->a);
	if (loc->w > 0) heading = PI - heading;
	heading *= 2;
	ConnUser[User].PointX = sin(heading);
	ConnUser[User].PointY = cos(heading);

	if (!ConnUser[User].Char)
		return;

	WORD OldLB = ConnUser[User].Char->Loc.landblock >> 16, NewLB = loc->landblock >> 16;
	
	memcpy(&ConnUser[User].Char->Loc, loc, sizeof(Location_t));

	DWORD tpf2 = 0x0000F748; memcpy (&MOVE_ENTITY[0x00],&tpf2,4);
	memcpy(&MOVE_ENTITY[0x4], &ConnUser[User].Char->GUID, 4);
	tpf2 = 0x00000034; memcpy (&MOVE_ENTITY[0x08],&tpf2,4);
	memcpy(&MOVE_ENTITY[0x0C], &ConnUser[User].Char->Loc, 4*5);
	memcpy(&MOVE_ENTITY[0x20], &ConnUser[User].Char->Loc.w, 4);

	if (Portal)
		ConnUser[User].PortalCount++;
	if (Override)
		ConnUser[User].OverrideCount++;
	
	memcpy (&MOVE_ENTITY[0x24],&ConnUser[User].Char->NumLogins,2);
	memcpy (&MOVE_ENTITY[0x26],&ConnUser[User].move_count,2);
	memcpy (&MOVE_ENTITY[0x28],&ConnUser[User].PortalCount,2);
	memcpy (&MOVE_ENTITY[0x2A],&ConnUser[User].OverrideCount,2);

	for (int iii=0;iii<MaxUsers;iii++)
	{
		if ((ConnUser[iii].Connected) && ((ConnUser[iii].State == 6) || (iii == User)))
			Send200(&MOVE_ENTITY[0], sizeof(MOVE_ENTITY), 3, iii);
	}
	ConnUser[User].move_count++;

	if (OldLB != NewLB)
	{
		LBObjects[OldLB].erase(ConnUser[User].Char->GUID);
		LBObjects[NewLB][ConnUser[User].Char->GUID] = (cWorldObject *) ConnUser[User].Char;


		if (OldLB == NewLB - 1) {
			SendLandblock(User, NewLB - 1);
			SendLandblock(User, NewLB - 1 - 256);
			SendLandblock(User, NewLB - 1 + 256);
		} else if (OldLB == NewLB + 1) {
			SendLandblock(User, NewLB + 1);
			SendLandblock(User, NewLB + 1 - 256);
			SendLandblock(User, NewLB + 1 + 256);
		} else if (OldLB == NewLB - 256) {
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256);
			SendLandblock(User, NewLB - 256 + 1);
		} else if (OldLB == NewLB + 256) {
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256);
			SendLandblock(User, NewLB + 256 + 1);
//Diagonals Now
		} else if (OldLB == NewLB + 256 + 1) {
			SendLandblock(User, NewLB + 256 + 1);
			SendLandblock(User, NewLB + 256 + 1 - 256);
			SendLandblock(User, NewLB + 256 + 1 - 256 - 256 );
			SendLandblock(User, NewLB + 256 + 1 - 1);
			SendLandblock(User, NewLB + 256 + 1 - 1 - 1);
		} else if (OldLB == NewLB + 256 - 1) {
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256 - 1 - 256);
			SendLandblock(User, NewLB + 256 - 1 - 256 - 256 );
			SendLandblock(User, NewLB + 256 - 1 + 1);
			SendLandblock(User, NewLB + 256 - 1 + 1 + 1);
		} else if (OldLB == NewLB - 256 + 1) {
			SendLandblock(User, NewLB - 256 + 1);
			SendLandblock(User, NewLB - 256 + 1 + 256);
			SendLandblock(User, NewLB - 256 + 1 + 256 + 256 );
			SendLandblock(User, NewLB - 256 + 1 - 1);
			SendLandblock(User, NewLB - 256 + 1 - 1 - 1);
		} else if (OldLB == NewLB - 256 - 1) {
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256 - 1 + 256);
			SendLandblock(User, NewLB - 256 - 1 + 256 + 256 );
			SendLandblock(User, NewLB - 256 - 1 + 1);
			SendLandblock(User, NewLB - 256 - 1 + 1 + 1);
		} else {
			SendLandblock(User, NewLB);
			SendLandblock(User, NewLB - 1);
			SendLandblock(User, NewLB + 1);
			SendLandblock(User, NewLB - 256);
			SendLandblock(User, NewLB - 256 - 1);
			SendLandblock(User, NewLB - 256 + 1);
			SendLandblock(User, NewLB + 256);
			SendLandblock(User, NewLB + 256 - 1);
			SendLandblock(User, NewLB + 256 + 1);
		}
	}
}

void cACServer::SendAnim(int User, WORD AnimNum)
{
	ConnUser[User].AnimCount++;

	cAnimStruct ASt;
	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0;
	ASt.wAnimation_Family = 0x3D;
	ASt.wUnk2 = 0x80;
	ASt.wUnk3 = 0;
	ASt.wAnimation_Seqnum = ConnUser[User].AnimCount;
	ASt.wAnimation_to_Play = (WORD) AnimNum;
	ASt.fPlaySpeed = 1.0f;

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ASt, sizeof(ASt), 3, i);
	}
}

void cACServer::PointAt(int User, DWORD GUIDAt)
{
	ConnUser[User].AnimCount++;

	cAnimStructEx ASt;
	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0;
	ASt.wAnimation_Family = 0x3D;

	ASt.dwf74c = 0x0000F74C;
	ASt.dwGUID = ConnUser[User].Char->GUID;
	ASt.wNumLogins = ConnUser[User].Char->NumLogins;
	ASt.wNumInteractions = ConnUser[User].move_count;
	ASt.wNumAnimations = ConnUser[User].AnimCount;
	ASt.wFlag = 0;
	ASt.wUnk1 = 0x0008;
	ASt.wAnimation_Family = 0x3D;
	ASt.GUIDTarget = GUIDAt;
	ASt.dwUnk1 = 0x429DB3C7;
	ASt.dwUnk2 = 0x0155EE0F;
	ASt.fUnk1 = 1.0f;
	ASt.Zero = 0;

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ASt, sizeof(ASt), 3, i);
	}

}

void cACServer::SendParticle(int User, DWORD ParticleNum)
{
	DWORD ParticleEffect[] = {
		0xF755,
		ConnUser[User].Char->GUID,
		ParticleNum,
		0
	};

	float tpf = 1.0f;
	memcpy(&ParticleEffect[3],&tpf,4);

	ConnUser[User].move_count++;

	for (int i=0;i<MaxUsers;i++)
	{
		if (ConnUser[i].Connected)
			Send200((BYTE *) &ParticleEffect, sizeof(ParticleEffect), 3, i);
	}
}

void cACServer::GiveXP(int User, int Type, DWORD Amount)
{
	BYTE Pack237[0xD];
	DWORD tpdword;	BYTE tpbyte;

	if ((User > 128) || (!ConnUser[User].Connected) || (ConnUser[User].State != 6))
	{
		ServerMessage(User, "Error in !givexp format... Make sure you're giving a valid User #...", COLOR_RED);
		
		return;
	}

	ConnUser[User].Char->TotalXP += Amount;

	ZeroMemory(Pack237, 0xD);
	tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
	tpbyte = ConnUser[User].Count237[0x15];	memcpy(&Pack237[4], &tpbyte, 1);
	tpdword = 0x15;	 memcpy(&Pack237[5], &tpdword, 4);		//Total XP
	tpdword = ConnUser[User].Char->TotalXP;  memcpy(&Pack237[9], &tpdword, 4);
	Send200(Pack237, sizeof(Pack237), 4, User);
	ConnUser[User].Count237[0x15]++;

	int NewLevel = 0;
	for (int i=ConnUser[User].Char->Level;i<128;i++)
	{
		if (ConnUser[User].Char->TotalXP < LevelArray[i])
		{
			NewLevel = i - 1;

			i = 128;
		}
	}
	if (NewLevel > ConnUser[User].Char->Level)
	{
		//Level him!
		SendParticle(User, 0x89);
		
		//Give skill credits as appropriate...
		for (i=ConnUser[User].Char->Level+1;i<=NewLevel;i++)
		{
			for (int h=0;h<=sizeof(CreditLevel)/4;h++)
				if (CreditLevel[h] == i)
					ConnUser[User].Char->SkillCredits++;
		}
		
		ConnUser[User].Char->Level = NewLevel;
		
		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x19];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x19;	 memcpy(&Pack237[5], &tpdword, 4);		//Level
		tpdword = ConnUser[User].Char->Level;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x19]++;

		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x18];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x18;	 memcpy(&Pack237[5], &tpdword, 4);		//Credits
		tpdword = ConnUser[User].Char->SkillCredits;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x18]++;

		sprintf(sayit, "You are now level %i!", ConnUser[User].Char->Level);
		ServerMessage(User, sayit, COLOR_CYAN);
		char mults[2] = { 's', 0 };
		if (ConnUser[User].Char->SkillCredits == 1)
			mults[0] = 0;
		printf(sayit, "You have %u experience points and %i skill credit%s available to raise skills and attributes.", ConnUser[User].Char->UnassignedXP, ConnUser[User].Char->SkillCredits, mults);
		ServerMessage(User, sayit, COLOR_CYAN);
		int NextCredit = 0, i=0;
		while (!NextCredit)
		{
			if (ConnUser[User].Char->Level < CreditLevel[i])
				NextCredit = CreditLevel[i];
			
			i++;
		}
		sprintf(sayit, "You will earn another skill credit at level %i.", NextCredit);
		ServerMessage(User, sayit, COLOR_CYAN);
	}

	if ((Type < 0) || ((Type > 39) && (Type != 128)))
	{
		ServerMessage(User, "Error in !givexp format... Type should be from 0-39 or 128 for unassigned XP...", COLOR_RED);

		return;
	}

	if (Type == SKILL_UNASSIGNED)
	{
		ConnUser[User].Char->UnassignedXP += Amount;
		
		ZeroMemory(Pack237, 0xD);
		tpdword = 0x237; memcpy(&Pack237[0], &tpdword, 4);
		tpbyte = ConnUser[User].Count237[0x16];	memcpy(&Pack237[4], &tpbyte, 1);
		tpdword = 0x16;	 memcpy(&Pack237[5], &tpdword, 4);		//Unassigned XP
		tpdword = ConnUser[User].Char->UnassignedXP;  memcpy(&Pack237[9], &tpdword, 4);
		Send200(Pack237, sizeof(Pack237), 4, User);
		ConnUser[User].Count237[0x16]++;
	} else if (Type < 40) {
		ConnUser[User].Char->SkillXP[Type] += Amount;

		int NewSkillLevel = 0;
		for (int i=0;i<200;i++)
		{
			if (ConnUser[User].Char->SkillXP[Type] < SkillArray[i])
			{
				NewSkillLevel = i - 1;

				i = 200;
			}
		}
		if (NewSkillLevel > ConnUser[User].Char->SkillInc[Type])
		{
			ConnUser[User].Char->SkillInc[Type] = NewSkillLevel;
		}

		//Send skill xp packet
	}
}

void cACServer::ActionComplete(int User)
{
	DWORD ActComp[] = {
		0x0000F7B0,
		ConnUser[User].Char->GUID,
		ConnUser[User].EventCount,
		0x000001C7,
		0x00000000
	};
	ConnUser[User].EventCount++;
	Send200((BYTE *) ActComp, sizeof(ActComp), 3, User);
}

float GetDistance(Location_t *loc1, Location_t *loc2)
{
	DWORD landblock;
	int XX, ns, ew;
	double xc, yc, PosNSa, PosEWa, PosNSb, PosEWb;

	float Distance;

	if ((loc1->landblock & 0x0000FF00) && (loc2->landblock & 0x0000FF00))
	{
		//Interior
		Distance = fabs(sqrt(pow(loc2->x - loc1->x,2) + pow(loc2->y - loc1->y,2) +  + pow(loc2->z - loc1->z,2)))/240;
	} else {
		//Exterior
		landblock = loc1->landblock;
		xc = loc1->x;
		yc = loc1->y;

		XX = landblock & 0x3F; 
		ns = ((((landblock & 0x00FF0000) >> 16) + 1) * 8) + ((XX - 1) & 7) - 1027;
		ew = ((((landblock & 0xFF000000) >> 24) + 1) * 8) + (int) ((XX - 1) / 8) - 1027;
    
		xc = xc - ((int) (xc / 24.0f) * 24.0f);
		yc = yc - ((int) (yc / 24.0f) * 24.0f);
		PosEWa = ((float) (ew) / 10.0f) + (xc / 240.0f);
		PosNSa = ((float) (ns) / 10.0f) + (yc / 240.0f);
		
		landblock = loc2->landblock;
		xc = loc2->x;
		yc = loc2->y;

		XX = landblock & 0x3F; 
		ns = ((((landblock & 0x00FF0000) >> 16) + 1) * 8) + ((XX - 1) & 7) - 1027;
		ew = ((((landblock & 0xFF000000) >> 24) + 1) * 8) + (int) ((XX - 1) / 8) - 1027;

		xc = xc - (float) ((int) (xc / 24.0f) * 24.0f);
		yc = yc - (float) ((int) (yc / 24.0f) * 24.0f);
		PosEWb = ((float) (ew) / 10.0f) + (xc / 240.0f);
		PosNSb = ((float) (ns) / 10.0f) + (yc / 240.0f);

		Distance = fabs(sqrt(pow(PosNSa - PosNSb,2) + pow(PosEWa - PosEWb,2) + pow((loc1->z - loc2->z)/240,2)));
	}

	return(Distance);
}

void cACServer::RemoveItem(DWORD GUID)
{
	DWORD F747[] = {
		0x0000F747,
		GUID
	};
	for (int i=0;i<MaxUsers;i++)
	{
		if ((ConnUser[i].Connected) && (ConnUser[i].State == 6))
			Send200((BYTE *) F747, sizeof(F747), 3, i);
	}
}

void cACServer::EvalObject(int User, DWORD GUID)
{
	cWorldObject *cwoItem = ObjectList[GUID];

	BYTE OutPack[0x500];
	BYTE *PackPointer = OutPack;
	DWORD tpd;//	WORD tpw;
	
	//F7b0 header
	tpd = 0xF7B0;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].Char->GUID, 4);	PackPointer += 4;
						memcpy(PackPointer, &ConnUser[User].EventCount, 4);	PackPointer += 4;
	tpd = 0x00C9;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//C9 = Identify Object
	ConnUser[User].EventCount++;

	//Figure out Flags...	
	switch (cwoItem->Type)
	{
	case OBJECT_CHARACTER:
		tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Flags - Char/Monster
		break;
	case OBJECT_PORTAL:
		tpd = 0x00010000;	memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Flags - Portal
		break;
	}

	tpd = 0x0001;		memcpy(PackPointer, &tpd, 4);	PackPointer += 4;		//Successful Assess

	//Build Packet
	if (cwoItem->Type == OBJECT_CHARACTER)
	{
		
	}
	else if (cwoItem->Type == OBJECT_PORTAL)
	{
		cWorldObject_Portal *tpp = (cWorldObject_Portal *) cwoItem;
		DWORD Lower, Upper;
		Lower = (tpp->Flags & 0x00FF0000) >> 16;
		if (!Lower) Lower = 0xFFFFFFFF;
		Upper = (tpp->Flags & 0xFF000000) >> 24;
		if (!Upper) Upper = 0xFFFFFFFF;

		memcpy(PackPointer, &Lower, 4);	PackPointer += 4;		//Lower Limit
		memcpy(PackPointer, &Upper, 4);	PackPointer += 4;		//Upper Limit

		int packlen;
		GenString(tpp->Name, &packlen, PackPointer);  PackPointer += packlen;
	}
	
	Send200(OutPack, (DWORD) PackPointer - (DWORD) OutPack, 3, User);
}